from typing import Type, Iterator, Tuple
import inspect

from arcor2.services import Service
import arcor2.services
from arcor2.data.services import ServiceTypeMetaDict, ServiceTypeMeta


def meta_from_def(type_def: Type[Service], built_in: bool = False) -> ServiceTypeMeta:

    return ServiceTypeMeta(type_def.__name__,
                           type_def.description(),
                           type_def.get_configuration_ids(),
                           built_in)


def built_in_services() -> Iterator[Tuple[str, Type[Service]]]:
    """
    Yields class name and class definition tuple
    """

    for cls in inspect.getmembers(arcor2.services, inspect.isclass):
        # ignore abstract services
        if not issubclass(cls[1], Service) or inspect.isabstract(cls[1]) or cls[1] == Service:
            continue

        yield cls[0], cls[1]


def get_built_in_service(name: str) -> Type[Service]:

    for bname, cls in built_in_services():
        if name == bname:
            return cls

    raise KeyError


def built_in_services_meta() -> ServiceTypeMetaDict:

    service_types: ServiceTypeMetaDict = {}

    # built-in object types
    for type_name, type_def in built_in_services():
        service_types[type_name] = meta_from_def(type_def, built_in=True)

    return service_types



from typing import Set, Optional
import time

from arcor2.services import Service
from arcor2.data.common import ActionMetadata
from arcor2.action import action


class TimeService(Service):
    """
    Time-related actions.
    """

    def __init__(self, configuration_id: str):  # TODO avoid need for configuration_id?
        super(TimeService, self).__init__(configuration_id)
        self._rate_start_time: Optional[float] = None

    @staticmethod
    def get_configuration_ids() -> Set[str]:
        return {"default"}

    @action
    def sleep(self, seconds: float = 1.0) -> None:

        assert 0.0 <= seconds <= 10.0e6

        time.sleep(seconds)

    @action
    def rate(self, period: float = 1.0) -> None:
        """
        Can be used to maintain desired rate of the loop. Should be the first action.
        :param period: Desired period in seconds.
        :return:
        """

        assert 0.0 <= period <= 10.0e6

        now = time.monotonic()

        if self._rate_start_time is None:
            self._rate_start_time = now
            return

        dif = self._rate_start_time + period - now

        if dif > 0:
            time.sleep(dif)

    sleep.__action__ = ActionMetadata(blocking=True)  # type: ignore
    rate.__action__ = ActionMetadata(blocking=True)  # type: ignore



# -*- coding: utf-8 -*-

from typing import List, Optional
import re

from arcor2.data.common import ActionState, ProjectState, CurrentAction
from dataclasses import dataclass, field
from dataclasses_jsonschema import JsonSchemaMixin

from arcor2.data.common import Scene, Project


def wo_suffix(name: str) -> str:
    return re.sub('Event$', '', name)


"""
------------------------------------------------------------------------------------------------------------------------
Common stuff
------------------------------------------------------------------------------------------------------------------------
"""


@dataclass
class Event(JsonSchemaMixin):

    event: str = field(default="", init=False)


"""
------------------------------------------------------------------------------------------------------------------------
Project / scene
------------------------------------------------------------------------------------------------------------------------
"""


@dataclass
class SceneChangedEvent(Event):

    data: Optional[Scene] = None
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ProjectChangedEvent(Event):

    data: Optional[Project] = None
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


"""
------------------------------------------------------------------------------------------------------------------------
Project execution
------------------------------------------------------------------------------------------------------------------------
"""


@dataclass
class ProjectExceptionEventData(JsonSchemaMixin):

    message: str = field(default_factory=str)
    type: str = field(default_factory=str)
    handled: bool = False


@dataclass
class ProjectExceptionEvent(Event):

    data: ProjectExceptionEventData = field(default_factory=ProjectExceptionEventData)
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class CurrentActionEvent(Event):

    data: CurrentAction = field(default_factory=CurrentAction)
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class ProjectStateEvent(Event):

    data: ProjectState = field(default_factory=ProjectState)
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class ActionStateEvent(Event):

    data: ActionState = field(default_factory=ActionState)
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ActionResult(JsonSchemaMixin):

    action_id: str = ""
    result: Optional[str] = field(default=None, metadata=dict(description="JSON-encoded result of the action."))
    error: Optional[str] = None


@dataclass
class ActionResultEvent(Event):

    data: ActionResult = field(default_factory=ActionResult)
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


"""
------------------------------------------------------------------------------------------------------------------------
Objects
------------------------------------------------------------------------------------------------------------------------
"""


@dataclass
class ObjectTypesChangedEvent(Event):

    data: List[str] = field(default_factory=list)  # changed object types
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821



from arcor2.data.rpc import common
from arcor2.data.rpc import execution
from arcor2.data.rpc import objects
from arcor2.data.rpc import robot
from arcor2.data.rpc import scene_project
from arcor2.data.rpc import services
from arcor2.data.rpc import storage

__all__ = ["common", "execution", "objects", "robot", "scene_project", "services", "storage"]



from typing import Set, Optional, Dict

from dataclasses import dataclass, field
from dataclasses_jsonschema import JsonSchemaMixin


@dataclass
class ServiceType(JsonSchemaMixin):
    """
    Service type, as it is stored in DB.
    """

    id: str
    source: str
    desc: Optional[str] = ""


@dataclass
class ServiceTypeMeta(JsonSchemaMixin):
    """
    Metadata about object type, as it is used in server API.
    """

    type: str
    description: str = field(default_factory=str)
    configuration_ids: Set[str] = field(default_factory=set)
    built_in: bool = False
    disabled: bool = False
    problem: Optional[str] = None


ServiceTypeMetaDict = Dict[str, ServiceTypeMeta]



from arcor2.nodes import build
from arcor2.nodes import execution
from arcor2.nodes import execution_proxy
from arcor2.nodes import server

__all__ = ["build", "execution", "execution_proxy", "server"]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import List, Any, Iterator, Optional, Tuple, Set, Dict, Union
from enum import Enum, unique
from uuid import UUID, uuid4

from json import JSONEncoder
from dataclasses import dataclass, field

import numpy as np  # type: ignore
import quaternion  # type: ignore

from dataclasses_jsonschema import JsonSchemaMixin

from arcor2.exceptions import Arcor2Exception


@unique
class StrEnum(str, Enum):

    @classmethod
    def set(cls) -> Set[str]:
        return set(map(lambda c: c.value, cls))  # type: ignore


@unique
class IntEnum(int, Enum):

    @classmethod
    def set(cls) -> Set[int]:
        return set(map(lambda c: c.value, cls))  # type: ignore


class ActionIOEnum(StrEnum):

    FIRST: str = "start"
    LAST: str = "end"


class DataClassEncoder(JSONEncoder):

    def default(self, o: Any) -> Any:

        if isinstance(o, JsonSchemaMixin):
            return o.to_dict()

        return super().default(o)


@dataclass
class IterableIndexable(JsonSchemaMixin):

    def __getitem__(self, item: int) -> float:

        attr = getattr(self, tuple(self.__dict__.keys())[item])
        assert isinstance(attr, float)
        return attr

    def __iter__(self) -> Iterator[float]:

        yield from self.__dict__.values()


@dataclass
class Position(IterableIndexable):

    x: float = 0.0
    y: float = 0.0
    z: float = 0.0


@dataclass
class Orientation(IterableIndexable):

    x: float = 0.0
    y: float = 0.0
    z: float = 0.0
    w: float = 1.0

    def as_quaternion(self) -> quaternion.quaternion:

        return np.quaternion(self.x, self.y, self.z, self.w)

    def set_from_quaternion(self, q: quaternion.quaternion) -> None:

        arr = quaternion.as_float_array(q)
        self.x = arr[0]
        self.y = arr[1]
        self.z = arr[2]
        self.w = arr[3]


@dataclass
class NamedOrientation(JsonSchemaMixin):

    id: str
    orientation: Orientation


@dataclass
class Pose(JsonSchemaMixin):

    position: Position = field(default_factory=Position)
    orientation: Orientation = field(default_factory=Orientation)


@dataclass
class ActionMetadata(JsonSchemaMixin):

    free: bool = False
    blocking: bool = False
    composite: bool = False
    blackbox: bool = False


@dataclass
class Joint(JsonSchemaMixin):

    name: str
    value: float


@dataclass
class ProjectRobotJoints(JsonSchemaMixin):

    id: str
    robot_id: str
    joints: List[Joint]
    is_valid: bool = False


@dataclass
class ActionPoint(JsonSchemaMixin):

    id: str
    position: Position
    orientations: List[NamedOrientation] = field(default_factory=list)
    robot_joints: List[ProjectRobotJoints] = field(default_factory=list)

    def pose(self, orientation_id: str = "default") -> Pose:

        for ori in self.orientations:
            if ori.id == orientation_id:
                return Pose(self.position, ori.orientation)
        raise Arcor2Exception(f"Action point {self.id} does not contain orientation {orientation_id}.")

    def get_joints(self, robot_id: str, joints_id: str) -> ProjectRobotJoints:

        for joints in self.robot_joints:
            if joints.id == joints_id and robot_id == joints.robot_id:
                return joints
        raise Arcor2Exception(f"Action point {self.id} does not contain robot joints {joints_id}.")


@dataclass
class SceneObject(JsonSchemaMixin):

    id: str
    type: str
    pose: Pose
    uuid: Optional[UUID] = None

    def __post_init__(self):

        if not self.uuid:
            self.uuid = uuid4()


@dataclass
class SceneService(JsonSchemaMixin):

    type: str
    configuration_id: str
    uuid: Optional[UUID] = None

    def __post_init__(self):

        if not self.uuid:
            self.uuid = uuid4()


@dataclass
class Scene(JsonSchemaMixin):

    id: str
    objects: List[SceneObject] = field(default_factory=list)
    services: List[SceneService] = field(default_factory=list)
    desc: str = field(default_factory=str)

    def __post_init__(self):

        self._id_cache: Dict[str, Union[SceneObject, SceneService]] = {}

    def object_or_service(self, object_or_service_id: str) -> Union[SceneObject, SceneService]:

        if not self._id_cache:
            for obj in self.objects:
                self._id_cache[obj.id] = obj
            for srv in self.services:
                self._id_cache[srv.type] = srv

        try:
            return self._id_cache[object_or_service_id]
        except KeyError:
            raise Arcor2Exception(f"Unknown object/service id: {object_or_service_id}, "
                                  f"known are: {self._id_cache.keys()}")  # TODO DataException?


@dataclass
class IdValue(JsonSchemaMixin):

    id: str
    value: str


class ActionParameterException(Arcor2Exception):
    pass


@dataclass
class ActionParameter(IdValue):

    type: str


@dataclass
class ActionIO(JsonSchemaMixin):

    default: str


@dataclass
class Action(JsonSchemaMixin):

    id: str
    type: str
    parameters: List[ActionParameter] = field(default_factory=list)
    inputs: List[ActionIO] = field(default_factory=list)
    outputs: List[ActionIO] = field(default_factory=list)
    uuid: Optional[UUID] = None

    def __post_init__(self):

        if not self.uuid:
            self.uuid = uuid4()

    def parse_type(self) -> Tuple[str, str]:

        try:
            obj_id, action = self.type.split("/")
        except ValueError:
            raise Arcor2Exception(f"Action: {self.id} has invalid type: {self.type}.")
        return obj_id, action

    def parameter(self, parameter_id) -> ActionParameter:

        for param in self.parameters:
            if parameter_id == param.id:
                return param

        raise Arcor2Exception("Param not found")


@dataclass
class ProjectActionPoint(ActionPoint):

    actions: List[Action] = field(default_factory=list)
    uuid: Optional[UUID] = None

    def __post_init__(self):

        if not self.uuid:
            self.uuid = uuid4()


@dataclass
class ProjectObject(JsonSchemaMixin):

    id: str
    action_points: List[ProjectActionPoint] = field(default_factory=list)
    uuid: Optional[UUID] = None


@dataclass
class Project(JsonSchemaMixin):

    id: str
    scene_id: str
    objects: List[ProjectObject] = field(default_factory=list)
    desc: str = field(default_factory=str)
    has_logic: bool = True

    def __post_init__(self):

        self._action_cache: Dict[str, Action] = {}
        self._ap_cache: Dict[str, ActionPoint] = {}

    def _build_cache(self):

        for obj in self.objects:
            for aps in obj.action_points:
                assert aps.id not in self._ap_cache, "Action point ID not unique."
                self._ap_cache[aps.id] = aps
                for act in aps.actions:
                    assert act.id not in self._action_cache, "Action ID not unique."
                    self._action_cache[act.id] = act

    def invalidate_cache(self):

        self._action_cache.clear()
        self._ap_cache.clear()

    def action(self, action_id: str) -> Action:

        if not self._action_cache:
            self._build_cache()

        try:
            return self._action_cache[action_id]
        except KeyError:
            raise Arcor2Exception("Action not found")

    def action_point(self, action_point_id: str) -> ActionPoint:

        if not self._ap_cache:
            self._build_cache()

        try:
            return self._ap_cache[action_point_id]
        except KeyError:
            raise Arcor2Exception("Action point not found")


@dataclass
class ProjectSources(JsonSchemaMixin):

    id: str  # project_id
    resources: str
    script: str


@dataclass
class IdDesc(JsonSchemaMixin):
    id: str
    desc: Optional[str] = None


@dataclass
class IdDescList(JsonSchemaMixin):

    items: List[IdDesc] = field(default_factory=list)


class ProjectStateEnum(Enum):

    RUNNING: str = "running"
    STOPPED: str = "stopped"
    PAUSED: str = "paused"


class ActionStateEnum(Enum):

    BEFORE: str = "before"
    AFTER: str = "after"


@dataclass
class ActionState(JsonSchemaMixin):

    object_id: str = ""
    method: str = ""
    where: ActionStateEnum = ActionStateEnum.BEFORE


@dataclass
class ProjectState(JsonSchemaMixin):

    state: ProjectStateEnum = ProjectStateEnum.STOPPED


@dataclass
class CurrentAction(JsonSchemaMixin):

    action_id: str = ""
    args: List[ActionParameter] = field(default_factory=list)


@dataclass
class BroadcastInfo(JsonSchemaMixin):

    host: str
    port: int



# -*- coding: utf-8 -*-

from typing import List, Set

from dataclasses import dataclass, field
from dataclasses_jsonschema import JsonSchemaMixin

from arcor2.data.common import IdDesc, SceneObject, SceneService
from arcor2.data.rpc.common import IdArgs, Request, Response, wo_suffix, TypeArgs


@dataclass
class SceneObjectUsageRequest(Request):

    args: IdArgs = field(metadata=dict(description="ID could be object or service."))
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class SceneObjectUsageResponse(Response):

    data: Set[str] = field(default_factory=set)
    response: str = field(default=SceneObjectUsageRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class AutoAddObjectToSceneRequest(Request):

    args: TypeArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class AutoAddObjectToSceneResponse(Response):

    response: str = field(default=AutoAddObjectToSceneRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------

@dataclass
class AddObjectToSceneRequest(Request):

    args: SceneObject
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class AddObjectToSceneResponse(Response):

    response: str = field(default=AddObjectToSceneRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class AddServiceToSceneRequest(Request):

    args: SceneService
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class AddServiceToSceneResponse(Response):

    response: str = field(default=AddServiceToSceneRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class RemoveFromSceneRequest(Request):

    args: IdArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class RemoveFromSceneResponse(Response):
    response: str = field(default=RemoveFromSceneRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class SaveSceneRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class SaveSceneResponse(Response):

    response: str = field(default=SaveSceneRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class SaveProjectRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class SaveProjectResponse(Response):

    response: str = field(default=SaveProjectRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class OpenProjectRequest(Request):

    args: IdArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class OpenProjectResponse(Response):

    response: str = field(default=OpenProjectRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class OpenSceneRequest(Request):

    args: IdArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class OpenSceneResponse(Response):

    response: str = field(default=OpenSceneRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class ListProjectsResponseData(IdDesc):

    valid: bool = field(default=False, metadata=dict(description="Objects and their actions exists."))
    executable: bool = field(default=False, metadata=dict(description="Logic is defined and valid."))
    problems: List[str] = field(default_factory=list)


@dataclass
class ListProjectsRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ListProjectsResponse(Response):

    data: List[ListProjectsResponseData] = field(default_factory=list)
    response: str = field(default=ListProjectsRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class ListScenesRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ListScenesResponse(Response):

    data: List[IdDesc] = field(default_factory=list)
    response: str = field(default=ListScenesRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class ExecuteActionArgs(JsonSchemaMixin):

    action_id: str = field(metadata=dict(description="ID of the action to be executed."))


@dataclass
class ExecuteActionRequest(Request):

    args: ExecuteActionArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ExecuteActionResponse(Response):

    response: str = field(default=ExecuteActionRequest.request, init=False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import json
import uuid
import tempfile
import base64
import os
import shutil

from apispec import APISpec  # type: ignore
from apispec_webframeworks.flask import FlaskPlugin  # type: ignore
from flask import Flask, jsonify, request, send_file
from flask_swagger_ui import get_swaggerui_blueprint  # type: ignore
import websocket  # type: ignore
from werkzeug.utils import secure_filename

import arcor2
from arcor2.data import rpc
from arcor2.data.helpers import RPC_MAPPING
from arcor2.nodes.execution import PORT as MANAGER_PORT
from arcor2.settings import PROJECT_PATH

PORT = 5009
SERVICE_NAME = "ARCOR2 Execution Service Proxy"

# Create an APISpec
spec = APISpec(
    title=SERVICE_NAME,
    version=arcor2.version(),
    openapi_version="3.0.2",
    plugins=[FlaskPlugin()],
)

app = Flask(__name__)


# TODO find out why long-lasting connection is not possible (it is getting disconnected after few minutes)
class WebsocketContextManager:

    def __init__(self):
        self.ws = None

    def __enter__(self):
        self.ws = websocket.create_connection(f"ws://0.0.0.0:{MANAGER_PORT}")
        return self.ws

    def __exit__(self, ex_type, ex_value, traceback):
        self.ws.close()
        if ex_type:
            raise ex_type(ex_value)


def call_rpc(req: rpc.common.Request) -> rpc.common.Response:

    with WebsocketContextManager() as ws:

        ws.send(req.to_json())

        while True:

            resp_dict = json.loads(ws.recv())
            if "response" not in resp_dict:  # ignore everything except rpc responses
                continue

            resp = RPC_MAPPING[resp_dict["response"]][1].from_dict(resp_dict)
            assert req.id == resp.id
            return resp


def get_id() -> int:
    return uuid.uuid4().int


def allowed_file(filename):
    return '.' in filename and \
           filename.rsplit('.', 1)[1].lower() == "zip"


@app.route("/packages/<string:package_id>", methods=['PUT'])
def put_package(package_id: str):
    """Put package
        ---
        put:
            description: Upload/update execution package.
            parameters:
                - in: path
                  name: package_id
                  schema:
                    type: string
                  required: true
                  description: unique ID
            requestBody:
                  content:
                    multipart/form-data:
                      schema:
                        type: object
                        properties:
                          # 'file' will be the field name in this multipart request
                          file:
                            type: string
                            format: binary
            responses:
                200:
                  description: Ok
                501:
                  description: Contains array of errors.
                  content:
                    application/json:
                      schema:
                        type: array
                        items:
                          type: string

    """

    file = request.files['file']
    file_name = secure_filename(file.filename)

    with tempfile.TemporaryDirectory() as tmpdirname:
        file_path = os.path.join(tmpdirname, file_name)
        file.save(file_path)

        with open(file_path, "rb") as zip_file:
            b64_bytes = base64.b64encode(zip_file.read())
            b64_str = b64_bytes.decode()

    resp = call_rpc(rpc.execution.UploadPackageRequest(
            id=get_id(),
            args=rpc.execution.UploadPackageArgs(package_id, b64_str)))

    if resp.result:
        return "ok", 200
    else:
        return jsonify(resp.messages), 501


@app.route("/packages/<string:package_id>", methods=['GET'])
def get_package(package_id: str):
    """Get execution package.
                ---
                get:
                  description: Get zip file with execution package.
                  parameters:
                    - in: path
                      name: package_id
                      schema:
                        type: string
                      required: true
                      description: unique ID
                  responses:
                    200:
                      description: Ok
                      content:
                        application/zip:
                            schema:
                              type: string
                              format: binary
                              example: The archive of execution package (.zip)
                    404:
                        description: Package ID was not found.
                """

    package_path = os.path.join(PROJECT_PATH, package_id)

    if not os.path.exists(package_path):
        return "Not found", 404

    with tempfile.TemporaryDirectory() as tmpdirname:

        archive_path = os.path.join(tmpdirname, package_id)
        shutil.make_archive(archive_path, 'zip', package_path)

        return send_file(archive_path + ".zip", as_attachment=True, cache_timeout=0)


# TODO use DataClassPlugin / PackageSummary model
@app.route("/packages", methods=['GET'])
def get_packages():
    """Gets summary for all stored execution packages.
                ---
                get:
                  description: Summary.
                  responses:
                    200:
                      description: Ok
                      content:
                        application/json:
                          schema:
                            type: array
                            items:
                              type: object
                              properties:
                                id:
                                  type: string
                                date_time:
                                  type: string
                                  format: date-time
                """

    resp = call_rpc(rpc.execution.ListPackagesRequest(id=get_id()))
    assert isinstance(resp, rpc.execution.ListPackagesResponse)

    return jsonify([d.to_dict() for d in resp.data]), 200


@app.route("/packages/<string:package_id>", methods=['DELETE'])
def delete_package(package_id: str):
    """Delete package.
            ---
            delete:
              description: Delete package.
              parameters:
                - in: path
                  name: package_id
                  schema:
                    type: string
                  required: true
                  description: unique ID
              responses:
                200:
                  description: Ok
                501:
                    description: Contains array of errors.
                    content:
                      application/json:
                        schema:
                          type: array
                          items:
                            type: string
            """

    resp = call_rpc(rpc.execution.DeletePackageRequest(id=get_id(), args=rpc.common.IdArgs(id=package_id)))

    if resp.result:
        return "ok", 200
    else:
        return jsonify(resp.messages), 501


@app.route("/packages/<string:package_id>/start", methods=['PUT'])
def package_start(package_id: str):
    """Run project
            ---
            put:
              description: Start execution of the execution package.
              parameters:
                - in: path
                  name: package_id
                  schema:
                    type: string
                  required: true
                  description: unique ID
              responses:
                200:
                  description: Ok
                501:
                    description: Contains array of errors.
                    content:
                      application/json:
                        schema:
                          type: array
                          items:
                            type: string
            """

    resp = call_rpc(rpc.execution.RunProjectRequest(id=get_id(), args=rpc.common.IdArgs(id=package_id)))

    if resp.result:
        return "ok", 200
    else:
        return jsonify(resp.messages), 501


@app.route("/packages/stop", methods=['PUT'])
def packages_stop():
    """Stops running project
            ---
            put:
              description: Stops execution of the given package.
              responses:
                200:
                  description: Ok
                501:
                    description: Contains array of errors.
                    content:
                      application/json:
                        schema:
                          type: array
                          items:
                            type: string
            """

    resp = call_rpc(rpc.execution.StopProjectRequest(id=get_id()))

    if resp.result:
        return "ok", 200
    else:
        return jsonify(resp.messages), 501


@app.route("/packages/pause", methods=['PUT'])
def packages_pause():
    """Pauses running package.
            ---
            put:
              description: Pause execution of the given package.
              responses:
                200:
                  description: Ok
                501:
                    description: Contains array of errors.
                    content:
                      application/json:
                        schema:
                          type: array
                          items:
                            type: string
            """

    resp = call_rpc(rpc.execution.PauseProjectRequest(id=get_id()))

    if resp.result:
        return "ok", 200
    else:
        return jsonify(resp.messages), 501


@app.route("/packages/resume", methods=['PUT'])
def packages_resume():
    """Resumes running package.
            ---
            put:
              description: Resumes execution of the given package.
              responses:
                200:
                  description: Ok
                501:
                    description: Contains array of errors.
                    content:
                      application/json:
                        schema:
                          type: array
                          items:
                            type: string
            """

    resp = call_rpc(rpc.execution.ResumeProjectRequest(id=get_id()))

    if resp.result:
        return "ok", 200
    else:
        return jsonify(resp.messages), 501


@app.route("/packages/active", methods=['GET'])
def packages_active():
    """Get id of the running package.
            ---
            get:
              description: Get id of the running package.
              responses:
                200:
                  description: Ok
                  content:
                    text/plain:
                      schema:
                        type: string
                        example: demo1
                404:
                    description: No project running
            """

    resp = call_rpc(rpc.execution.ProjectStateRequest(id=get_id()))
    assert isinstance(resp, rpc.execution.ProjectStateResponse)

    if resp.data.id:
        return resp.data.id, 200
    else:
        return "No project running.", 404


@app.route("/swagger/api/swagger.json", methods=["GET"])
def get_swagger():
    return json.dumps(spec.to_dict())


with app.test_request_context():
    spec.path(view=put_package)
    spec.path(view=get_package)
    spec.path(view=get_packages)
    spec.path(view=delete_package)
    spec.path(view=package_start)
    spec.path(view=packages_stop)
    spec.path(view=packages_pause)
    spec.path(view=packages_resume)
    spec.path(view=packages_active)


def main():

    parser = argparse.ArgumentParser(description=SERVICE_NAME)
    parser.add_argument('-s', '--swagger', action="store_true", default=False)
    args = parser.parse_args()

    if args.swagger:
        print(spec.to_yaml())
        return

    SWAGGER_URL = "/swagger"

    swaggerui_blueprint = get_swaggerui_blueprint(
        SWAGGER_URL,  # Swagger UI static files will be mapped to '{SWAGGER_URL}/dist/'
        f"./api/swagger.json"
    )

    # Register blueprint at URL
    app.register_blueprint(swaggerui_blueprint, url_prefix=SWAGGER_URL)

    app.run(host='0.0.0.0', port=PORT)


if __name__ == '__main__':
    main()



import copy
import inspect
from typing import Dict, Iterator, Tuple, Type, Set, get_type_hints, Union

import horast

import arcor2
from arcor2.data.object_type import ObjectTypeMetaDict, ObjectActionsDict, ObjectTypeMeta, ActionParameterMeta, \
    ObjectAction, ObjectActions
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types import Generic
from arcor2.services import Service
from arcor2.docstring import parse_docstring
from arcor2.parameter_plugins.base import ParameterPlugin, ParameterPluginException
from arcor2.source.utils import find_function
from arcor2.service_types_utils import built_in_services

SERVICES_METHOD_NAME = "from_services"


class ObjectTypeException(Arcor2Exception):
    pass


def built_in_types() -> Iterator[Tuple[str, Type[Generic]]]:
    """
    Yields class name and class definition tuple
    """

    for cls in inspect.getmembers(arcor2.object_types, inspect.isclass):
        if not issubclass(cls[1], Generic):
            continue

        yield cls[0], cls[1]


def get_built_in_type(name: str) -> Type[Generic]:

    for bname, cls in built_in_types():
        if name == bname:
            return cls

    raise KeyError


def built_in_types_names() -> Set[str]:

    names = set()

    for type_name, _ in built_in_types():
        names.add(type_name)

    return names


class DataError(Arcor2Exception):
    pass


def obj_description_from_base(data: Dict[str, ObjectTypeMeta], obj_type: ObjectTypeMeta) -> str:

    try:
        obj = data[obj_type.base]
    except KeyError:
        raise DataError(f"Unknown object type: {obj_type}.")

    if obj.description:
        return obj.description

    if not obj.base:
        return ""

    return obj_description_from_base(data, data[obj.base])


def built_in_types_meta() -> ObjectTypeMetaDict:

    d: ObjectTypeMetaDict = {}

    # built-in object types
    for type_name, type_def in built_in_types():
        d[type_name] = meta_from_def(type_def, built_in=True)

    return d


def meta_from_def(type_def: Type[Generic], built_in: bool = False) -> ObjectTypeMeta:

    obj = ObjectTypeMeta(type_def.__name__,
                         type_def.description(),
                         built_in=built_in,
                         # TODO kind of hack to make Generic abstract
                         abstract=inspect.isabstract(type_def) or type_def == Generic)

    bases = inspect.getmro(type_def)

    if len(bases) > 2:
        obj.base = bases[1].__name__

    if hasattr(type_def, SERVICES_METHOD_NAME):

        srv_method = getattr(type_def, SERVICES_METHOD_NAME)

        if not inspect.isfunction(srv_method):
            raise ObjectTypeException(f"{SERVICES_METHOD_NAME} should be method.")

        for name, ttype in get_type_hints(srv_method).items():
            if name == "return":
                continue
            obj.needs_services.add(ttype.__name__)

    # TODO check whether the same services are also in constructor

    return obj


def built_in_types_actions(plugins: Dict[Type, Type[ParameterPlugin]]) -> ObjectActionsDict:

    # TODO get built_in_types sources and rather use get_object_actions(source)?

    d: ObjectActionsDict = {}

    # built-in object types / services
    for type_name, type_def in (*built_in_types(), *built_in_services()):

        assert issubclass(type_def, (Generic, Service))

        if type_name not in d:
            d[type_name] = []
        d[type_name] = object_actions(plugins, type_def, inspect.getsource(type_def))

    return d


class IgnoreActionException(Arcor2Exception):
    pass


def object_actions(plugins: Dict[Type, Type[ParameterPlugin]], type_def: Union[Type[Generic], Type[Service]],
                   source: str) -> ObjectActions:

    ret: ObjectActions = []
    tree = horast.parse(source)

    # ...inspect.ismethod does not work on un-initialized classes
    for method_name, method_def in inspect.getmembers(type_def, predicate=inspect.isfunction):

        # TODO check also if the method has 'action' decorator (ast needed)
        if not hasattr(method_def, "__action__"):
            continue

        # action from ancestor, will be copied later (only if the action was not overridden)
        base_cls_def = type_def.__bases__[0]
        if hasattr(base_cls_def, method_name) and getattr(base_cls_def, method_name) == method_def:
            continue

        meta = method_def.__action__

        data = ObjectAction(name=method_name, meta=meta)

        doc = parse_docstring(method_def.__doc__)
        doc_short = doc["short_description"]
        if doc_short:
            data.description = doc_short

        signature = inspect.signature(method_def)

        method_tree = find_function(method_name, tree)

        try:
            for name, ttype in get_type_hints(method_def).items():

                try:
                    param_type = plugins[ttype]
                except KeyError:
                    for k, v in plugins.items():
                        if not v.EXACT_TYPE and inspect.isclass(ttype) and issubclass(ttype, k):
                            param_type = v
                            break
                    else:
                        if name == "return" and ttype == type(None):  # noqa: E721
                            # ...just ignore NoneType for returns
                            continue

                        # ignore action with unknown parameter type
                        raise IgnoreActionException(f"Parameter {name} of action {method_name}"
                                                    f" has unknown type {ttype}.")

                if name == "return":
                    data.returns = param_type.type_name()
                    continue

                args = ActionParameterMeta(name=name, type=param_type.type_name())
                try:
                    param_type.meta(args, method_def, method_tree)
                except ParameterPluginException as e:
                    # TODO log exception
                    raise IgnoreActionException(e)

                if name in type_def.DYNAMIC_PARAMS:
                    args.dynamic_value = True
                    dvp = type_def.DYNAMIC_PARAMS[name][1]
                    if dvp:
                        args.dynamic_value_parents = dvp

                def_val = signature.parameters[name].default
                if def_val is not inspect.Parameter.empty:
                    args.default_value = def_val

                try:
                    args.description = doc["params"][name].strip()
                except KeyError:
                    pass

                data.parameters.append(args)

        except IgnoreActionException as e:
            data.disabled = True
            data.problem = str(e)
            # TODO log exception

        ret.append(data)

    return ret


def add_ancestor_actions(obj_type_name: str,
                         object_actions: ObjectActionsDict,
                         object_types: ObjectTypeMetaDict) -> None:

    base_name = object_types[obj_type_name].base

    if not base_name:
        return

    if object_types[base_name].base:
        add_ancestor_actions(base_name, object_actions, object_types)

    # do not add action from base if it is overridden in child
    for base_action in object_actions[base_name]:
        for obj_action in object_actions[obj_type_name]:
            if base_action.name == obj_action.name:

                # built-in object has no "origins" yet
                if not obj_action.origins:
                    obj_action.origins = base_name
                break
        else:
            action = copy.deepcopy(base_action)
            if not action.origins:
                action.origins = base_name
            object_actions[obj_type_name].append(action)



from typing import Union
from typing_extensions import Literal

from typed_ast.ast3 import Assign, Name, ClassDef, Str, Module, \
    ImportFrom, alias, Pass, AnnAssign, Store, Load, Subscript, Index

from arcor2.data.object_type import ObjectTypeMeta
from arcor2.helpers import camel_case_to_snake_case
from arcor2.source import SourceException
from arcor2.source.utils import get_name, tree_to_str, find_function, get_name_attr
from arcor2.object_types_utils import built_in_types_names, meta_from_def, object_actions
import arcor2.helpers as hlp
import arcor2.object_types
from arcor2.object_types import Generic
from arcor2.parameter_plugins import TYPE_TO_PLUGIN


def check_object_type(object_type_source: str, type_name: str) -> None:
    """
    Checks whether the object type source is a valid one.
    :param object_type_source:
    :return:
    """
    try:
        type_def = hlp.type_def_from_source(object_type_source, type_name, Generic)
    except hlp.TypeDefException as e:
        raise SourceException(e)

    meta_from_def(type_def, False)
    object_actions(TYPE_TO_PLUGIN, type_def, object_type_source)


def fix_object_name(object_id: str) -> str:

    return camel_case_to_snake_case(object_id).replace(' ', '_')


# TODO could this be done like this https://stackoverflow.com/a/9269964/3142796 ??
def new_object_type_source(parent: ObjectTypeMeta, child: ObjectTypeMeta) -> str:

    assert parent.type == child.base

    tree = Module(body=[], type_ignores=[])

    if parent.type in built_in_types_names():
        import_from = arcor2.object_types.__name__
    else:
        import_from = camel_case_to_snake_case(parent.type)

    tree.body.append(ImportFrom(module=import_from,
                                names=[alias(name=parent.type, asname=None)],
                                level=0))

    c = ClassDef(name=child.type,
                 bases=[get_name(parent.type)],
                 keywords=[],
                 body=[],
                 decorator_list=[])

    # TODO add docstring with description (if provided)
    c.body.append(Pass())

    tree.body.append(c)

    return tree_to_str(tree)


def object_instance_from_res(tree: Module, object_id: str, cls_name: str,
                             cls_type: Union[Literal["objects"], Literal["services"]]) -> None:

    main_body = find_function("main", tree).body
    last_assign_idx = -1

    for body_idx, body_item in enumerate(main_body):

        if isinstance(body_item, (Assign, AnnAssign)):
            last_assign_idx = body_idx

    assign = AnnAssign(
        target=Name(
            id=fix_object_name(object_id),
            ctx=Store()),
        annotation=Name(
            id=cls_name,
            ctx=Load()),
        value=Subscript(
            value=get_name_attr('res', cls_type),
            slice=Index(value=Str(
                s=object_id,
                kind='')),
            ctx=Load()),
        simple=1)

    main_body.insert(last_assign_idx + 1, assign)



from typing import List
import json

from arcor2.data.common import Project, Pose, Scene
from arcor2.parameter_plugins.base import ParameterPlugin, TypesDict
from arcor2.parameter_plugins.list import ListParameterPlugin, get_type_name


class PosePlugin(ParameterPlugin):

    @classmethod
    def type(cls):
        return Pose

    @classmethod
    def value(cls, type_defs: TypesDict, scene: Scene, project: Project, action_id: str, parameter_id: str) -> Pose:

        obj_id, ap_id, value_id = cls.parse_id(project.action(action_id).parameter(parameter_id))
        return project.action_point(ap_id).pose(value_id)

    @classmethod
    def value_to_json(cls, value: Pose) -> str:
        return value.to_json()


class PoseListPlugin(ListParameterPlugin):

    @classmethod
    def type(cls):
        return List[Pose]

    @classmethod
    def type_name(cls) -> str:
        return get_type_name(PosePlugin)

    @classmethod
    def value(cls, type_defs: TypesDict, scene: Scene, project: Project, action_id: str, parameter_id: str) \
            -> List[Pose]:

        ret: List[Pose] = []

        for obj_id, ap_id, value_id in cls.parse_list_id(project.action(action_id).parameter(parameter_id)):
            ret.append(project.action_point(ap_id).pose(value_id))

        return ret

    @classmethod
    def value_to_json(cls, value: List[Pose]) -> str:
        return json.dumps([v.to_json() for v in value])



# -*- coding: utf-8 -*-

from typing import List, Optional
from datetime import datetime

from dataclasses import dataclass, field
from dataclasses_jsonschema import JsonSchemaMixin

from arcor2.data.common import ProjectState, ActionState, CurrentAction
from arcor2.data.rpc.common import IdArgs, Request, Response, wo_suffix


@dataclass
class BuildProjectRequest(Request):
    """
    Calls Build service to generate execution package and uploads it to the Execution service.
    """

    args: IdArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class BuildProjectResponse(Response):

    response: str = field(default=BuildProjectRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class UploadPackageArgs(JsonSchemaMixin):

    id: str = field(metadata=dict(description="Id of the execution package (so far == ID of the project)."))
    data: str = field(metadata=dict(description="Base64 encoded content of the zip file."))


@dataclass
class UploadPackageRequest(Request):

    args: UploadPackageArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class UploadPackageResponse(Response):

    response: str = field(default=UploadPackageRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class ListPackagesRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ModifiedFile(JsonSchemaMixin):

    filename: str
    modified: datetime


@dataclass
class PackageSummary(JsonSchemaMixin):

    id: str
    date_time: datetime
    # modified_files: List[ModifiedFile] = field(default_factory=list)


@dataclass
class ListPackagesResponse(Response):

    data: List[PackageSummary] = field(default_factory=list)
    response: str = field(default=ListPackagesRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------

@dataclass
class DeletePackageRequest(Request):

    args: IdArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class DeletePackageResponse(Response):

    response: str = field(default=DeletePackageRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class RunProjectRequest(Request):

    args: IdArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class RunProjectResponse(Response):

    response: str = field(default=RunProjectRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class StopProjectRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class StopProjectResponse(Response):

    response: str = field(default=StopProjectRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class PauseProjectRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class PauseProjectResponse(Response):

    response: str = field(default=PauseProjectRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------

@dataclass
class ProjectStateRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ProjectStateData(JsonSchemaMixin):

    id: Optional[str] = None
    project: ProjectState = field(default_factory=ProjectState)
    action: Optional[ActionState] = None
    action_args: Optional[CurrentAction] = None


@dataclass
class ProjectStateResponse(Response):

    data: ProjectStateData = field(default_factory=ProjectStateData)
    response: str = field(default=ProjectStateRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class ResumeProjectRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ResumeProjectResponse(Response):

    response: str = field(default=ResumeProjectRequest.request, init=False)



import select
import sys
from typing import Union, Callable, Any, TYPE_CHECKING, no_type_check
from arcor2.data.events import Event, ProjectStateEvent, ActionStateEvent
from arcor2.data.common import ProjectStateEnum, ActionStateEnum, ActionState, ProjectState
from functools import wraps

HANDLE_ACTIONS = True

if TYPE_CHECKING:
    from arcor2.object_types import Generic  # NOQA
    from arcor2.services import Service  # NOQA


def read_stdin(timeout: float = 0.0) -> Union[str, None]:

    if select.select([sys.stdin], [], [], timeout)[0]:
        return sys.stdin.readline().strip()
    return None


def handle_action(inst: Union["Service", "Generic"], f: Callable[..., Any], where: ActionStateEnum) -> None:

    # can't import Service/Generic here (circ. import)
    if hasattr(inst, "id"):
        obj_id = inst.id  # type: ignore
    else:
        obj_id = inst.__class__.__name__

    print_event(ActionStateEvent(data=ActionState(obj_id, f.__name__, where)))

    ctrl_cmd = read_stdin()

    if ctrl_cmd == "p":
        print_event(ProjectStateEvent(data=ProjectState(ProjectStateEnum.PAUSED)))
        while True:
            ctrl_cmd = read_stdin(0.1)
            if ctrl_cmd == "r":
                print_event(ProjectStateEvent(data=ProjectState(ProjectStateEnum.RUNNING)))
                break


def print_event(event: Event) -> None:
    """
    Used from main script to print event as JSON.
    """

    print(event.to_json())
    sys.stdout.flush()


@no_type_check
def action(f):

    @wraps(f)
    def wrapper(*args: Union["Generic", Any], **kwargs: Any) -> Any:

        # automagical overload for dictionary (allow to get rid of ** in script).
        if len(args) == 2 and isinstance(args[1], dict) and not kwargs:
            kwargs = args[1]
            args = (args[0],)

        if not action.inside_composite and HANDLE_ACTIONS:
            handle_action(args[0], f, ActionStateEnum.BEFORE)

        if wrapper.__action__.composite:  # TODO and not step_into
            action.inside_composite = f

        res = f(*args, **kwargs)

        if action.inside_composite == f:
            action.inside_composite = None

        if not action.inside_composite and HANDLE_ACTIONS:
            handle_action(args[0], f, ActionStateEnum.AFTER)

        return res

    return wrapper


action.inside_composite = None  # type: ignore



from typing import Tuple, Dict, Union

from arcor2.data.common import Project, Action, ActionIOEnum, ProjectObject, ActionPoint
from arcor2.helpers import camel_case_to_snake_case
from arcor2.exceptions import ActionPointNotFound, Arcor2Exception


class ProjectException(Arcor2Exception):
    pass


def get_actions_cache(project: Project) -> Tuple[Dict[str, Action], Union[str, None], Union[str, None]]:

    actions_cache = {}
    first_action_id = None
    last_action_id = None

    for obj in project.objects:
        for aps in obj.action_points:
            for act in aps.actions:
                actions_cache[act.id] = act
                if act.inputs and act.inputs[0].default == ActionIOEnum.FIRST.value:
                    if first_action_id is not None:
                        raise ProjectException("Multiple starts.")
                    first_action_id = act.id
                elif act.outputs and act.outputs[0].default == ActionIOEnum.LAST.value:
                    if last_action_id is not None:
                        raise ProjectException("Multiple ends.")
                    last_action_id = act.id

    if first_action_id not in actions_cache:
        raise ProjectException(f"Unknown start action: {first_action_id}.")

    if last_action_id not in actions_cache:
        raise ProjectException(f"Unknown end action: {last_action_id}.")

    return actions_cache, first_action_id, last_action_id


def get_objects_cache(project: Project, id_to_var: bool = False) -> Dict[str, ProjectObject]:

    cache: Dict[str, ProjectObject] = {}

    for obj in project.objects:
        if id_to_var:
            cache[camel_case_to_snake_case(obj.id)] = obj
        else:
            cache[obj.id] = obj

    return cache


def clear_project_logic(project: Project) -> None:

    for obj in project.objects:
        for act_point in obj.action_points:
            for action in act_point.actions:
                action.inputs.clear()
                action.outputs.clear()


def get_object_ap(project: Project, ap_id: str) -> Tuple[ProjectObject, ActionPoint]:

    for obj in project.objects:
        for ap in obj.action_points:
            if ap.id == ap_id:
                return obj, ap

    raise ActionPointNotFound



from typing import List

from arcor2.object_types import Generic
from arcor2.action import action
from arcor2.data.common import ActionMetadata


class Tester(Generic):
    """
    A generic tester
    """

    @action
    def run_test(self, seq: List[float], seq_id: List[str], eqp_res: List[bool]) -> List[bool]:
        return [bool(seq), bool(seq_id), bool(eqp_res)]

    run_test.__action__ = ActionMetadata(blocking=True)



import arcor2.helpers as hlp
from arcor2.source import SourceException
from arcor2.object_types_utils import object_actions
from arcor2.services import Service
from arcor2.parameter_plugins import TYPE_TO_PLUGIN


def check_service_type(service_type_source: str, type_name: str) -> None:
    """
    Checks whether the service type source is a valid one.
    :param service_type_source: Source code.
    :param type_name: Class name.
    :return:
    """
    try:
        type_def = hlp.type_def_from_source(service_type_source, type_name, Service)
    except hlp.TypeDefException as e:
        raise SourceException(e)

    # meta_from_def(type_def)  # calls API which is probably undesirable for check
    object_actions(TYPE_TO_PLUGIN, type_def, service_type_source)



import inspect
from typing import Optional

from arcor2.source.object_types import check_object_type
from arcor2.source import SourceException
from arcor2.helpers import import_cls, ImportClsException
from arcor2.exceptions import Arcor2Exception
from arcor2 import persistent_storage as storage
from arcor2.data.object_type import ObjectType, Models
from arcor2.object_types import Generic


class UploadException(Arcor2Exception):
    pass


def upload(module_cls: str, model: Optional[Models] = None) -> None:

    try:
        _, cls = import_cls(module_cls)
    except ImportClsException as e:
        raise UploadException(e)

    path = inspect.getfile(cls)

    with open(path, "r") as source_file:

        source = source_file.read()

        try:
            check_object_type(source, cls.__name__)
        except SourceException as e:
            print(e)
            raise UploadException(f"There is something wrong with source code of '{cls.__name__}'.")

        assert issubclass(cls, Generic)

        obj_type = ObjectType(id=cls.__name__, source=source, desc=cls.description())

        if model:
            obj_type.model = model.metamodel()
            storage.put_model(model)

        print(f"Storing '{obj_type.id}'...")
        storage.update_object_type(obj_type)



import traceback
import time
import sys
from typing import Optional, Dict, Callable, Tuple, Type, Union, Any, Awaitable, TypeVar
from types import ModuleType
import json
import asyncio
import importlib
import re
import keyword
import logging

from dataclasses_jsonschema import ValidationError

import websockets
from aiologger.formatters.base import Formatter  # type: ignore

from arcor2.data.rpc.common import Request
from arcor2.data.events import Event, ProjectExceptionEvent, ProjectExceptionEventData
from arcor2.data.helpers import RPC_MAPPING, EVENT_MAPPING
from arcor2.exceptions import Arcor2Exception
from arcor2.data.common import Pose, Position, Orientation


RPC_RETURN_TYPES = Union[None, Tuple[bool, str]]

# TODO what's wrong with following type?
# RPC_DICT_TYPE = Dict[Type[Request], Callable[[Request], Coroutine[Any, Any, Union[Response, RPC_RETURN_TYPES]]]]
RPC_DICT_TYPE = Dict[Type[Request], Any]

# TODO replace Any with WebsocketSomething
# EVENT_DICT_TYPE = Dict[Type[Event], Callable[[Any, Event], Coroutine[Any, Any, None]]]
EVENT_DICT_TYPE = Dict[Type[Event], Any]

LOG_FORMAT = '%(name)s - %(levelname)-8s: %(message)s'


class ImportClsException(Arcor2Exception):
    pass


class TypeDefException(Arcor2Exception):
    pass


def aiologger_formatter() -> Formatter:

    return Formatter(LOG_FORMAT)


def logger_formatter() -> logging.Formatter:

    return logging.Formatter(LOG_FORMAT)


def import_cls(module_cls: str) -> Tuple[ModuleType, Type[Any]]:
    """
    Gets module and class based on string like 'module/Cls'.
    :param module_cls:
    :return:
    """

    try:
        module_name, cls_name = module_cls.split('/')
    except (IndexError, ValueError):
        raise ImportClsException("Invalid format.")

    try:
        module = importlib.import_module(module_name)
    except ModuleNotFoundError:
        raise ImportClsException(f"Module '{module_name}' not found.")

    try:
        cls = getattr(module, cls_name)
    except AttributeError:
        raise ImportClsException(f"Class {cls_name} not found in module '{module_name}'.")

    return module, cls


def is_valid_identifier(value: str) -> bool:
    """
    Identifier (e.g. object id) will be used as variable name in the script - it should be in snake_case,
    not containing any special characters etc.
    :param value:
    :return:
    """

    return value.isidentifier() and not keyword.iskeyword(value) and value == camel_case_to_snake_case(value)


def is_valid_type(value: str) -> bool:
    """
    Value will be used as object type name - it should be in CamelCase,
    not containing any special characters etc.
    :param value:
    :return:
    """

    return value.isidentifier() and not keyword.iskeyword(value) and value == snake_case_to_camel_case(value)


_first_cap_re = re.compile('(.)([A-Z][a-z]+)')
_all_cap_re = re.compile('([a-z0-9])([A-Z])')


def camel_case_to_snake_case(camel_str: str) -> str:

    s1 = _first_cap_re.sub(r'\1_\2', camel_str)
    return _all_cap_re.sub(r'\1_\2', s1).lower()


def snake_case_to_camel_case(snake_str: str) -> str:

    return re.sub(r"(?:^|_)(.)", lambda m: m.group(1).upper(), snake_str)


async def server(client: Any,
                 path: str,
                 logger: Any,
                 register: Callable[[Any], Awaitable[None]],
                 unregister: Callable[[Any], Awaitable[None]],
                 rpc_dict: RPC_DICT_TYPE,
                 event_dict: Optional[EVENT_DICT_TYPE] = None) -> None:

    if event_dict is None:
        event_dict = {}

    await register(client)
    try:
        async for message in client:

            try:
                data = json.loads(message)
            except json.decoder.JSONDecodeError as e:
                await logger.error(f"Invalid data: '{message}'.")
                await logger.debug(e)
                continue

            if not isinstance(data, dict):
                await logger.error(f"Invalid data: '{data}'.")
                continue

            if "request" in data:  # ...then it is RPC

                try:
                    req_cls, resp_cls = RPC_MAPPING[data['request']]
                except KeyError:
                    await logger.error(f"Unknown RPC request: {data}.")
                    continue

                if req_cls not in rpc_dict:
                    await logger.debug(f"Ignoring RPC request: {data}.")
                    continue

                try:
                    req = req_cls.from_dict(data)
                except ValidationError as e:
                    await logger.error(f"Invalid RPC: {data}, error: {e}")
                    continue

                try:
                    resp = await rpc_dict[req_cls](req)
                except Arcor2Exception as e:
                    await logger.exception(e)
                    resp = False, "System error."

                if resp is None:  # default response
                    resp = resp_cls()
                elif isinstance(resp, tuple):
                    resp = resp_cls(result=resp[0], messages=[resp[1]])
                else:
                    assert isinstance(resp, resp_cls)

                resp.id = req.id

                await asyncio.wait([client.send(resp.to_json())])
                await logger.debug(f"RPC request: {req}, result: {resp}")

            elif "event" in data:  # ...event from UI

                try:
                    event_cls = EVENT_MAPPING[data["event"]]
                except KeyError as e:
                    await logger.error(f"Unknown event type: {e}.")
                    continue

                if event_cls not in event_dict:
                    await logger.debug(f"Ignoring event: {data}.")
                    continue

                try:
                    event = event_cls.from_dict(data)
                except ValidationError as e:
                    await logger.error(f"Invalid event: {data}, error: {e}")
                    continue

                await event_dict[event_cls](client, event)

            else:
                await logger.error(f"unsupported format of message: {data}")
    except websockets.exceptions.ConnectionClosed:
        pass
    finally:
        await unregister(client)


def format_stacktrace():
    parts = ["Traceback (most recent call last):\n"]
    parts.extend(traceback.format_stack(limit=25)[:-2])
    parts.extend(traceback.format_exception(*sys.exc_info())[1:])
    return "".join(parts)


def print_exception(e: Exception) -> None:

    pee = ProjectExceptionEvent(data=ProjectExceptionEventData(str(e),
                                                               e.__class__.__name__,
                                                               isinstance(e, Arcor2Exception)))
    print(pee.to_json())
    sys.stdout.flush()

    with open("traceback-{}.txt".format(time.strftime("%Y%m%d-%H%M%S")), "w") as tb_file:
        tb_file.write(format_stacktrace())


async def run_in_executor(func, *args):
    return await asyncio.get_event_loop().run_in_executor(None, func, *args)


def make_pose_rel(parent: Pose, child: Pose) -> Pose:
    """
    :param parent: e.g. scene object
    :param child:  e.g. action point
    :return: relative pose
    """

    p = Pose()

    p.position.x = child.position.x - parent.position.x
    p.position.y = child.position.y - parent.position.y
    p.position.z = child.position.z - parent.position.z

    p.orientation.set_from_quaternion(child.orientation.as_quaternion()/parent.orientation.as_quaternion())

    return p


def make_position_abs(parent: Position, child: Position) -> Position:

    p = Position()

    p.x = child.x + parent.x
    p.y = child.y + parent.y
    p.z = child.z + parent.z
    return p


def make_orientation_abs(parent: Orientation, child: Orientation) -> Orientation:

    p = Orientation()
    p.set_from_quaternion(child.as_quaternion() * parent.as_quaternion())
    return p


def make_pose_abs(parent: Pose, child: Pose) -> Pose:
    """
    :param parent: e.g. scene object
    :param child:  e.g. action point
    :return: absolute pose
    """

    p = Pose()
    p.position = make_position_abs(parent.position, child.position)
    p.orientation = make_orientation_abs(parent.orientation, child.orientation)
    return p


T = TypeVar('T')


def type_def_from_source(source: str, type_name: str, output_type: Type[T]) -> Type[T]:

    mod = ModuleType('temp_module')
    try:
        exec(source, mod.__dict__)
    except Exception as e:  # exec might raise almost anything
        raise TypeDefException(e)

    try:
        cls_def = getattr(mod, type_name)
    except AttributeError:
        raise TypeDefException(f"Source does not contain class named '{type_name}'.")

    if not issubclass(cls_def, output_type):
        raise TypeDefException(f"Class is not of expected type.")

    return cls_def



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import asyncio
import json
import functools
import sys
from typing import Union, Set, Optional
import os
import argparse
import tempfile
import base64
import zipfile
import shutil

import websockets
from websockets.server import WebSocketServerProtocol
from aiologger import Logger  # type: ignore
from aiologger.levels import LogLevel  # type: ignore
from dataclasses_jsonschema import ValidationError

import arcor2
from arcor2.helpers import server, aiologger_formatter, RPC_RETURN_TYPES, RPC_DICT_TYPE
from arcor2.source.utils import make_executable
from arcor2.settings import PROJECT_PATH
from arcor2.data import rpc
from arcor2.data.events import Event, ProjectStateEvent, ActionStateEvent, CurrentActionEvent
from arcor2.data.common import ProjectStateEnum, ProjectState
from arcor2.data.helpers import EVENT_MAPPING

PORT = 6790

logger = Logger.with_default_handlers(name='manager', formatter=aiologger_formatter())

PROCESS: Union[asyncio.subprocess.Process, None] = None
PROJECT_EVENT: ProjectStateEvent = ProjectStateEvent()
ACTION_EVENT: Optional[ActionStateEvent] = None
ACTION_ARGS_EVENT: Optional[CurrentActionEvent] = None
PROJECT_ID: Optional[str] = None
TASK = None

CLIENTS: Set = set()

MAIN_SCRIPT_NAME = "script.py"


def process_running() -> bool:

    return PROCESS is not None and PROCESS.returncode is None


async def project_state(event: ProjectStateEvent):

    global PROJECT_EVENT
    PROJECT_EVENT = event
    await send_to_clients(event)


async def read_proc_stdout() -> None:

    global PROJECT_EVENT
    global ACTION_EVENT
    global ACTION_ARGS_EVENT
    global PROJECT_ID

    logger.info("Reading script stdout...")

    assert PROCESS is not None
    assert PROCESS.stdout is not None

    await project_state(ProjectStateEvent(ProjectState(ProjectStateEnum.RUNNING)))

    while process_running():
        try:
            stdout = await PROCESS.stdout.readuntil()
        except asyncio.exceptions.IncompleteReadError:
            break

        stripped = stdout.decode("utf-8").strip()

        try:
            data = json.loads(stripped)
        except json.decoder.JSONDecodeError as e:
            await logger.error(f"Script printed out: {stripped}")
            await logger.debug(e)
            continue

        if not isinstance(data, dict) or "event" not in data:
            await logger.error("Strange data from script: {}".format(data))
            continue

        try:
            evt = EVENT_MAPPING[data["event"]].from_dict(data)
        except ValidationError as e:
            await logger.error("Invalid event: {}, error: {}".format(data, e))
            continue

        if isinstance(evt, ProjectStateEvent):
            await project_state(evt)
            continue
        elif isinstance(evt, ActionStateEvent):
            ACTION_EVENT = evt
        elif isinstance(evt, CurrentActionEvent):
            ACTION_ARGS_EVENT = evt

        await send_to_clients(evt)

    ACTION_EVENT = None
    ACTION_ARGS_EVENT = None
    PROJECT_ID = None

    await project_state(ProjectStateEvent(ProjectState(ProjectStateEnum.STOPPED)))

    logger.info(f"Process finished with returncode {PROCESS.returncode}.")


async def project_run(req: rpc.execution.RunProjectRequest) -> Union[rpc.execution.RunProjectResponse,
                                                                     RPC_RETURN_TYPES]:

    global PROCESS
    global TASK
    global PROJECT_ID

    if process_running():
        return False, "Already running!"

    package_path = os.path.join(PROJECT_PATH, req.args.id)

    try:
        os.chdir(package_path)
    except FileNotFoundError:
        return False, "Not found."

    script_path = os.path.join(package_path, MAIN_SCRIPT_NAME)

    try:
        make_executable(script_path)
    except FileNotFoundError:
        return False, "Not an execution package."

    await logger.info(f"Starting script: {script_path}")
    PROCESS = await asyncio.create_subprocess_exec(script_path, stdin=asyncio.subprocess.PIPE,
                                                   stdout=asyncio.subprocess.PIPE,
                                                   stderr=asyncio.subprocess.STDOUT)
    if PROCESS.returncode is not None:
        return False, "Failed to start project."
    PROJECT_ID = req.args.id
    TASK = asyncio.ensure_future(read_proc_stdout())  # run task in background


async def project_stop(req: rpc.execution.StopProjectRequest) -> Union[rpc.execution.StopProjectResponse,
                                                                       RPC_RETURN_TYPES]:

    if not process_running():
        return False, "Project not running."

    assert PROCESS is not None
    assert TASK is not None

    await logger.info("Terminating process")
    PROCESS.terminate()
    await logger.info("Waiting for process to finish...")
    await asyncio.wait([TASK])


async def project_pause(req: rpc.execution.PauseProjectRequest) -> Union[rpc.execution.PauseProjectResponse,
                                                                         RPC_RETURN_TYPES]:

    if not process_running():
        return False, "Project not running."

    assert PROCESS is not None
    assert PROCESS.stdin is not None

    if PROJECT_EVENT.data.state != ProjectStateEnum.RUNNING:
        return False, "Cannot pause."

    PROCESS.stdin.write("p\n".encode())
    await PROCESS.stdin.drain()
    return None


async def project_resume(req: rpc.execution.ResumeProjectRequest) -> Union[rpc.execution.ResumeProjectResponse,
                                                                           RPC_RETURN_TYPES]:

    if not process_running():
        return False, "Project not running."

    assert PROCESS is not None and PROCESS.stdin is not None

    if PROJECT_EVENT.data.state != ProjectStateEnum.PAUSED:
        return False, "Cannot resume."

    PROCESS.stdin.write("r\n".encode())
    await PROCESS.stdin.drain()
    return None


async def project_state_cb(req: rpc.execution.ProjectStateRequest) -> Union[rpc.execution.ProjectStateResponse,
                                                                            RPC_RETURN_TYPES]:

    resp = rpc.execution.ProjectStateResponse()
    resp.data.project = PROJECT_EVENT.data
    if ACTION_EVENT:
        resp.data.action = ACTION_EVENT.data
    if ACTION_ARGS_EVENT:
        resp.data.action_args = ACTION_ARGS_EVENT.data
    resp.data.id = PROJECT_ID
    return resp


async def _upload_package_cb(req: rpc.execution.UploadPackageRequest) -> Union[rpc.execution.UploadPackageResponse,
                                                                               RPC_RETURN_TYPES]:

    target_path = os.path.join(PROJECT_PATH, req.args.id)

    # TODO do not allow if there are manual changes?

    with tempfile.TemporaryDirectory() as tmpdirname:

        zip_path = os.path.join(tmpdirname, "publish.zip")

        b64_bytes = req.args.data.encode()
        zip_content = base64.b64decode(b64_bytes)

        with open(zip_path, "wb") as zip_file:
            zip_file.write(zip_content)

        try:
            with zipfile.ZipFile(zip_path, 'r') as zip_ref:
                zip_ref.extractall(tmpdirname)
        except zipfile.BadZipFile as e:
            await logger.error(e)
            return False, "Invalid zip file."

        os.remove(zip_path)

        try:
            shutil.rmtree(target_path)
        except FileNotFoundError:
            pass
        shutil.copytree(tmpdirname, target_path)

    script_path = os.path.join(target_path, MAIN_SCRIPT_NAME)
    make_executable(script_path)
    return None


async def list_packages_cb(req: rpc.execution.ListPackagesRequest) -> Union[rpc.execution.ListPackagesResponse,
                                                                            RPC_RETURN_TYPES]:

    resp = rpc.execution.ListPackagesResponse()

    subfolders = [f.path for f in os.scandir(PROJECT_PATH) if f.is_dir()]

    from datetime import datetime, timedelta, timezone

    for folder_path in subfolders:

        if not os.path.isfile(os.path.join(folder_path, MAIN_SCRIPT_NAME)):
            continue

        # TODO read actual project timestamp from data/project.json
        resp.data.append(rpc.execution.PackageSummary(os.path.basename(folder_path),
                                                      datetime.now(timezone.utc) - timedelta(hours=1)))

        # TODO report manual changes (check last modification of files)?

    return resp


async def delete_package_cb(req: rpc.execution.DeletePackageRequest) -> Union[rpc.execution.DeletePackageResponse,
                                                                              RPC_RETURN_TYPES]:

    if PROJECT_ID and PROJECT_ID == req.args.id:
        return False, "Package is being executed."

    target_path = os.path.join(PROJECT_PATH, req.args.id)

    try:
        shutil.rmtree(target_path)
    except FileNotFoundError:
        return False, "Not found."

    return None


async def send_to_clients(event: Event) -> None:

    if CLIENTS:
        data = event.to_json()
        await asyncio.wait([client.send(data) for client in CLIENTS])


async def register(websocket: WebSocketServerProtocol) -> None:

    await logger.info("Registering new client")
    CLIENTS.add(websocket)
    # TODO send current state


async def unregister(websocket: WebSocketServerProtocol) -> None:
    await logger.info("Unregistering client")
    CLIENTS.remove(websocket)

RPC_DICT: RPC_DICT_TYPE = {
    rpc.execution.RunProjectRequest: project_run,
    rpc.execution.StopProjectRequest: project_stop,
    rpc.execution.PauseProjectRequest: project_pause,
    rpc.execution.ResumeProjectRequest: project_resume,
    rpc.execution.ProjectStateRequest: project_state_cb,
    rpc.execution.UploadPackageRequest: _upload_package_cb,
    rpc.execution.ListPackagesRequest: list_packages_cb,
    rpc.execution.DeletePackageRequest: delete_package_cb
}


def main() -> None:

    assert sys.version_info >= (3, 8)

    parser = argparse.ArgumentParser()

    parser.add_argument("-v", "--verbose", help="Increase output verbosity",
                        action="store_const", const=LogLevel.DEBUG, default=LogLevel.INFO)
    parser.add_argument('--version', action='version', version=arcor2.version(),
                        help="Shows ARCOR2 version and exits.")
    parser.add_argument('--api_version', action='version', version=arcor2.api_version(),
                        help="Shows API version and exits.")
    parser.add_argument("-a", "--asyncio_debug", help="Turn on asyncio debug mode.",
                        action="store_const", const=True, default=False)

    args = parser.parse_args()
    logger.level = args.verbose

    loop = asyncio.get_event_loop()
    loop.set_debug(enabled=args.asyncio_debug)

    bound_handler = functools.partial(server, logger=logger, register=register, unregister=unregister,
                                      rpc_dict=RPC_DICT)
    loop.run_until_complete(
        websockets.serve(bound_handler, '0.0.0.0', PORT))
    loop.run_forever()


if __name__ == "__main__":
    main()



from typing import Dict, Set, Tuple
import warnings
import pkgutil

# key: name of parameter, value: name of method to call (to get set of strings), set of parent parameters
DynamicParamDict = Dict[str, Tuple[str, Set[str]]]

warnings.filterwarnings(
    action='ignore',
    category=UserWarning,
    module='quaternion'
)

warnings.filterwarnings(
    action='ignore',
    category=UserWarning,
    module='dataclasses_jsonschema',
    message="Unable to create schema for 'Any'"
)


def _version(file: str) -> str:

    res = pkgutil.get_data(__name__, file)
    if not res:
        return "unknown"
    return res.decode().strip()


def version() -> str:
    return _version('VERSION')


def api_version() -> str:
    return _version('API_VERSION')



# -*- coding: utf-8 -*-

from typing import List

from dataclasses import dataclass, field

from arcor2.data.robot import RobotMeta
from arcor2.data.rpc.common import Request, Response, wo_suffix


@dataclass
class GetRobotMetaRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class GetRobotMetaResponse(Response):

    data: List[RobotMeta] = field(default_factory=list)
    response: str = field(default=GetRobotMetaRequest.request, init=False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import asyncio
import base64
import json
import functools
import sys
from typing import Dict, Set, Union, TYPE_CHECKING, Tuple, Optional, List, Callable, cast, AsyncIterator, \
    get_type_hints, Type, Any
import uuid
import argparse
import os
import tempfile

import websockets
from websockets.server import WebSocketServerProtocol
from aiologger import Logger  # type: ignore
from aiologger.levels import LogLevel  # type: ignore

import arcor2
from arcor2.source.logic import program_src
from arcor2.source.object_types import new_object_type_source
from arcor2.source import SourceException
from arcor2 import nodes
from arcor2 import service_types_utils as stu, object_types_utils as otu, helpers as hlp
from arcor2.data.common import Scene, Project, Pose, Position, SceneObject, SceneService, ActionIOEnum,\
    Joint, NamedOrientation, ProjectRobotJoints
from arcor2.data.object_type import ObjectActionsDict, ObjectTypeMetaDict, Model3dType, \
    MeshFocusAction, ObjectModel, Models, ObjectTypeMeta
from arcor2.data.services import ServiceTypeMetaDict, ServiceTypeMeta
from arcor2.data.robot import RobotMeta
from arcor2.data import rpc, events
from arcor2.data.helpers import RPC_MAPPING
from arcor2.persistent_storage import PersistentStorageException
from arcor2 import aio_persistent_storage as storage
from arcor2.object_types import Generic, Robot
from arcor2.project_utils import get_object_ap
from arcor2.scene_utils import get_scene_object
from arcor2.exceptions import ActionPointNotFound, SceneObjectNotFound, Arcor2Exception
from arcor2.services import Service, RobotService
from arcor2.parameter_plugins import TYPE_TO_PLUGIN, PARAM_PLUGINS
from arcor2.parameter_plugins.base import TypesDict, ParameterPluginException
from arcor2 import action as action_mod
from arcor2 import rest

# disables before/after messages, etc.
action_mod.HANDLE_ACTIONS = False


if TYPE_CHECKING:
    ReqQueue = asyncio.Queue[rpc.common.Request]
    RespQueue = asyncio.Queue[rpc.common.Response]
else:
    ReqQueue = asyncio.Queue
    RespQueue = asyncio.Queue


logger = Logger.with_default_handlers(name='server', formatter=hlp.aiologger_formatter(), level=LogLevel.DEBUG)

MANAGER_URL = os.getenv("ARCOR2_EXECUTION_URL", f"ws://0.0.0.0:{nodes.execution.PORT}")
BUILDER_URL = os.getenv("ARCOR2_BUILDER_URL", f"http://0.0.0.0:{nodes.build.PORT}")

SCENE: Union[Scene, None] = None
PROJECT: Union[Project, None] = None

INTERFACES: Set[WebSocketServerProtocol] = set()

MANAGER_RPC_REQUEST_QUEUE: ReqQueue = ReqQueue()
MANAGER_RPC_RESPONSES: Dict[int, RespQueue] = {}

OBJECT_TYPES: ObjectTypeMetaDict = {}
SERVICE_TYPES: ServiceTypeMetaDict = {}
ROBOT_META: Dict[str, RobotMeta] = {}
TYPE_DEF_DICT: TypesDict = {}

# TODO merge it into one dict?
SCENE_OBJECT_INSTANCES: Dict[str, Generic] = {}
SERVICES_INSTANCES: Dict[str, Service] = {}

ACTIONS: ObjectActionsDict = {}  # used for actions of both object_types / services

FOCUS_OBJECT: Dict[str, Dict[int, Pose]] = {}  # object_id / idx, pose
FOCUS_OBJECT_ROBOT: Dict[str, rpc.common.RobotArg] = {}  # key: object_id


RUNNING_ACTION: Optional[str] = None

PORT: int = int(os.getenv("ARCOR2_SERVER_PORT", 6789))


class RobotPoseException(Arcor2Exception):
    pass


# TODO refactor into server_utils
def scene_needed(coro):
    @functools.wraps(coro)
    async def async_wrapper(*args, **kwargs):

        if SCENE is None or not SCENE.id:
            return False, "Scene not opened or has invalid id."
        return await coro(*args, **kwargs)

    return async_wrapper


def no_project(coro):
    @functools.wraps(coro)
    async def async_wrapper(*args, **kwargs):
        if PROJECT:
            return False, "Not available during project editing."
        return await coro(*args, **kwargs)

    return async_wrapper


def project_needed(coro):
    @functools.wraps(coro)
    async def async_wrapper(*args, **kwargs):

        if PROJECT is None or not PROJECT.id:
            return False, "Project not opened or has invalid id."
        return await coro(*args, **kwargs)

    return async_wrapper


async def handle_manager_incoming_messages(manager_client):

    try:

        async for message in manager_client:

            msg = json.loads(message)

            if "event" in msg and INTERFACES:
                await asyncio.wait([intf.send(message) for intf in INTERFACES])
            elif "response" in msg:

                # TODO handle potential errors
                _, resp_cls = RPC_MAPPING[msg["response"]]
                resp = resp_cls.from_dict(msg)
                MANAGER_RPC_RESPONSES[resp.id].put_nowait(resp)

    except websockets.exceptions.ConnectionClosed:
        await logger.error("Connection to manager closed.")


async def project_manager_client() -> None:

    while True:

        await logger.info("Attempting connection to manager...")

        try:

            async with websockets.connect(MANAGER_URL) as manager_client:

                await logger.info("Connected to manager.")

                future = asyncio.ensure_future(handle_manager_incoming_messages(manager_client))

                while True:

                    if future.done():
                        break

                    try:
                        msg = await asyncio.wait_for(MANAGER_RPC_REQUEST_QUEUE.get(), 1.0)
                    except asyncio.TimeoutError:
                        continue

                    try:
                        await manager_client.send(msg.to_json())
                    except websockets.exceptions.ConnectionClosed:
                        await MANAGER_RPC_REQUEST_QUEUE.put(msg)
                        break
        except ConnectionRefusedError as e:
            await logger.error(e)
            await asyncio.sleep(delay=1.0)


def scene_event() -> events.SceneChangedEvent:

    return events.SceneChangedEvent(SCENE)


def project_event() -> events.ProjectChangedEvent:

    return events.ProjectChangedEvent(PROJECT)


async def notify(event: events.Event, exclude_ui=None):

    if (exclude_ui is None and INTERFACES) or (exclude_ui and len(INTERFACES) > 1):
        message = event.to_json()
        await asyncio.wait([intf.send(message) for intf in INTERFACES if intf != exclude_ui])


async def _notify(interface, msg_source: Callable[[], events.Event]):

    await notify(msg_source(), interface)


async def notify_scene_change_to_others(interface: Optional[WebSocketServerProtocol] = None) -> None:

    await _notify(interface, scene_event)


async def notify_project_change_to_others(interface=None) -> None:

    await _notify(interface, project_event)


async def notify_scene(interface) -> None:
    message = scene_event().to_json()
    await asyncio.wait([interface.send(message)])


async def notify_project(interface) -> None:
    message = project_event().to_json()
    await asyncio.wait([interface.send(message)])


async def _initialize_server() -> None:

    while True:  # wait until Project service becomes available
        try:
            await storage.get_projects()
            break
        except storage.PersistentStorageException:
            await asyncio.sleep(1)

    # this has to be done sequentially as objects might depend on services so (all) services has to be known first
    await _get_service_types()
    await _get_object_types()

    await asyncio.wait([_get_object_actions(), _check_manager()])

    bound_handler = functools.partial(hlp.server, logger=logger, register=register, unregister=unregister,
                                      rpc_dict=RPC_DICT, event_dict=EVENT_DICT)

    await logger.info("Server initialized.")
    await asyncio.wait([websockets.serve(bound_handler, '0.0.0.0', PORT)])


async def _get_robot_meta(robot_type: Union[Type[Robot], Type[RobotService]]) -> None:

    meta = RobotMeta(robot_type.__name__)
    meta.features.focus = hasattr(robot_type, "focus")  # TODO more sophisticated test? (attr(s) and return value?)
    ROBOT_META[robot_type.__name__] = meta


async def _get_service_types() -> None:

    global SERVICE_TYPES

    service_types: ServiceTypeMetaDict = {}

    srv_ids = await storage.get_service_type_ids()

    for srv_id in srv_ids.items:

        srv_type = await storage.get_service_type(srv_id.id)

        try:
            type_def = hlp.type_def_from_source(srv_type.source, srv_type.id, Service)
            meta = stu.meta_from_def(type_def)
            service_types[srv_id.id] = meta
        except Arcor2Exception as e:
            await logger.exception(f"Disabling service type {srv_type.id}.")
            service_types[srv_id.id] = ServiceTypeMeta(srv_id.id, "Service not available.", disabled=True,
                                                       problem=str(e))
            continue

        if not meta.configuration_ids:
            meta.disabled = True
            meta.problem = "No configuration available."
            continue

        TYPE_DEF_DICT[srv_id.id] = type_def

        if issubclass(type_def, RobotService):
            asyncio.ensure_future(_get_robot_meta(type_def))

    SERVICE_TYPES = service_types


async def _get_object_types() -> None:

    global OBJECT_TYPES

    object_types: ObjectTypeMetaDict = otu.built_in_types_meta()

    obj_ids = await storage.get_object_type_ids()

    for obj_id in obj_ids.items:
        obj = await storage.get_object_type(obj_id.id)
        try:
            type_def = hlp.type_def_from_source(obj.source, obj.id, Generic)
            meta = otu.meta_from_def(type_def)
            object_types[obj.id] = meta
        except (otu.ObjectTypeException, hlp.TypeDefException) as e:
            await logger.exception(f"Disabling object type {obj.id}.")
            object_types[obj.id] = ObjectTypeMeta(obj_id.id, "Object type disabled.", disabled=True, problem=str(e))
            continue

        for srv in meta.needs_services:
            try:
                if SERVICE_TYPES[srv].disabled:
                    meta.disabled = True
                    meta.problem = f"Depends on disabled service '{srv}'."
                    break
            except KeyError:
                meta.disabled = True
                meta.problem = f"Depends on unknown service '{srv}'."
                break

        TYPE_DEF_DICT[obj.id] = type_def

        if obj.model:
            model = await storage.get_model(obj.model.id, obj.model.type)
            kwargs = {model.type().value.lower(): model}
            object_types[obj.id].object_model = ObjectModel(model.type(), **kwargs)  # type: ignore

        if issubclass(type_def, Robot):
            asyncio.ensure_future(_get_robot_meta(type_def))

    # if description is missing, try to get it from ancestor(s)
    for obj_type, obj_meta in object_types.items():
        if obj_meta.description:
            continue

        try:
            obj_meta.description = otu.obj_description_from_base(object_types, obj_meta)
        except otu.DataError as e:
            await logger.error(f"Failed to get info from base for {obj_type}, error: '{e}'.")

    OBJECT_TYPES = object_types


async def get_object_types_cb(req: rpc.objects.GetObjectTypesRequest) -> rpc.objects.GetObjectTypesResponse:
    return rpc.objects.GetObjectTypesResponse(data=list(OBJECT_TYPES.values()))


async def get_services_cb(req: rpc.services.GetServicesRequest) -> rpc.services.GetServicesResponse:
    return rpc.services.GetServicesResponse(data=list(SERVICE_TYPES.values()))


@scene_needed
@no_project
async def save_scene_cb(req: rpc.scene_project.SaveSceneRequest) -> Union[rpc.scene_project.SaveSceneResponse,
                                                                          hlp.RPC_RETURN_TYPES]:

    assert SCENE

    try:
        stored_scene = await storage.get_scene(SCENE.id)
    except storage.PersistentStorageException:
        # new scene, no need for further checks
        await storage.update_scene(SCENE)
        return None

    # let's check if something important has changed
    for old_obj in stored_scene.objects:
        for new_obj in SCENE.objects:
            if old_obj.id != new_obj.id:
                continue

            if old_obj.pose != new_obj.pose:
                asyncio.ensure_future(scene_object_pose_updated(SCENE.id, new_obj.id))

    await storage.update_scene(SCENE)
    return None


@scene_needed
@project_needed
async def save_project_cb(req: rpc.scene_project.SaveProjectRequest) -> Union[rpc.scene_project.SaveProjectResponse,
                                                                              hlp.RPC_RETURN_TYPES]:

    assert SCENE and PROJECT
    await storage.update_project(PROJECT)
    return None


async def clear_scene() -> None:

    global SCENE

    await logger.info("Clearing the scene.")
    rs = find_robot_service()
    if rs:
        for obj_inst in SCENE_OBJECT_INSTANCES.values():
            await collision(obj_inst, rs, remove=True)
    SCENE_OBJECT_INSTANCES.clear()
    SERVICES_INSTANCES.clear()  # TODO call destroy

    SCENE = None


async def open_scene(scene_id: str) -> None:

    global SCENE
    SCENE = await storage.get_scene(scene_id)

    for srv in SCENE.services:
        res, msg = await add_service_to_scene(srv)
        if not res:
            await clear_scene()
            raise Arcor2Exception(msg)

    for obj in SCENE.objects:
        res, msg = await add_object_to_scene(obj, add_to_scene=False, srv_obj_ok=True)
        if not res:
            await clear_scene()
            raise Arcor2Exception(msg)

    assert {srv.type for srv in SCENE.services} == SERVICES_INSTANCES.keys()
    assert {obj.id for obj in SCENE.objects} == SCENE_OBJECT_INSTANCES.keys()

    asyncio.ensure_future(notify_scene_change_to_others())


async def open_project(project_id: str) -> None:

    global PROJECT

    PROJECT = await storage.get_project(project_id)
    await open_scene(PROJECT.scene_id)

    assert SCENE
    for obj in PROJECT.objects:
        # TODO how to handle missing object?
        scene_obj = SCENE.object_or_service(obj.id)
        obj.uuid = scene_obj.uuid

    asyncio.ensure_future(notify_project_change_to_others())


@no_project
async def open_scene_cb(req: rpc.scene_project.OpenSceneRequest) -> Union[rpc.scene_project.OpenSceneResponse,
                                                                          hlp.RPC_RETURN_TYPES]:

    try:
        await open_scene(req.args.id)
    except Arcor2Exception as e:
        await logger.exception(f"Failed to open scene {req.args.id}.")
        return False, str(e)
    return None


async def open_project_cb(req: rpc.scene_project.OpenProjectRequest) -> Union[rpc.scene_project.OpenProjectResponse,
                                                                              hlp.RPC_RETURN_TYPES]:

    # TODO validate using project_problems?
    try:
        await open_project(req.args.id)
    except Arcor2Exception as e:
        await logger.exception(f"Failed to open project {req.args.id}.")
        return False, str(e)

    return None


def valid_object_types() -> ObjectTypeMetaDict:
    """
    To get only valid (not disabled) types.
    :return:
    """

    return {obj_type: obj for obj_type, obj in OBJECT_TYPES.items() if not obj.disabled}


def valid_service_types() -> ServiceTypeMetaDict:
    """
    To get only valid (not disabled) types.
    :return:
    """

    return {srv_type: srv for srv_type, srv in SERVICE_TYPES.items() if not srv.disabled}


async def _get_object_actions() -> None:  # TODO do it in parallel

    global ACTIONS

    object_actions_dict: ObjectActionsDict = otu.built_in_types_actions(TYPE_TO_PLUGIN)

    valid_types = valid_object_types()

    for obj_type, obj in valid_types.items():

        if obj.built_in:  # built-in types are already there
            continue

        # db-stored (user-created) object types
        obj_db = await storage.get_object_type(obj_type)
        try:
            type_def = hlp.type_def_from_source(obj_db.source, obj_db.id, Generic)
            object_actions_dict[obj_type] = otu.object_actions(TYPE_TO_PLUGIN, type_def, obj_db.source)
        except hlp.TypeDefException as e:
            await logger.error(e)

    # add actions from ancestors
    for obj_type in valid_types.keys():
        otu.add_ancestor_actions(obj_type, object_actions_dict, OBJECT_TYPES)

    # get services' actions
    for service_type, service_meta in valid_service_types().items():

        if service_meta.built_in:
            continue

        srv_type = await storage.get_service_type(service_type)
        try:
            srv_type_def = hlp.type_def_from_source(srv_type.source, service_type, Service)
            object_actions_dict[service_type] = otu.object_actions(TYPE_TO_PLUGIN, srv_type_def, srv_type.source)
        except Arcor2Exception:
            await logger.exception(f"Error while processing service type {service_type}")

    ACTIONS = object_actions_dict

    await notify(events.ObjectTypesChangedEvent(data=list(object_actions_dict.keys())))


async def _check_manager() -> None:
    """
    Loads project if it is loaded on manager
    :return:
    """

    # TODO avoid cast
    resp = cast(rpc.execution.ProjectStateResponse,
                await manager_request(rpc.execution.ProjectStateRequest(id=uuid.uuid4().int)))  # type: ignore

    if resp.data.id is not None and (PROJECT is None or PROJECT.id != resp.data.id):
        await open_project(resp.data.id)


async def get_object_actions_cb(req: rpc.objects.GetActionsRequest) -> Union[rpc.objects.GetActionsResponse,
                                                                             hlp.RPC_RETURN_TYPES]:

    try:
        return rpc.objects.GetActionsResponse(data=ACTIONS[req.args.type])
    except KeyError:
        return False, f"Unknown object type: '{req.args.type}'."


async def manager_request(req: rpc.common.Request) -> rpc.common.Response:

    assert req.id not in MANAGER_RPC_RESPONSES

    MANAGER_RPC_RESPONSES[req.id] = RespQueue(maxsize=1)
    await MANAGER_RPC_REQUEST_QUEUE.put(req)

    resp = await MANAGER_RPC_RESPONSES[req.id].get()
    del MANAGER_RPC_RESPONSES[req.id]
    return resp


async def get_robot_joints(robot_id: str) -> List[Joint]:
    """
    :param robot_id:
    :return: List of joints
    """

    robot_inst = await get_robot_instance(robot_id)

    if isinstance(robot_inst, Robot):

        try:
            return await hlp.run_in_executor(robot_inst.robot_joints)
        except NotImplementedError:
            raise RobotPoseException("The robot does not support getting joints.")

    elif isinstance(robot_inst, RobotService):

        try:
            return await hlp.run_in_executor(robot_inst.robot_joints, robot_id)
        except NotImplementedError:
            raise RobotPoseException("The robot does not support getting joints.")

    else:
        raise Arcor2Exception("Not a robot instance.")


async def get_end_effector_pose(robot_id: str, end_effector: str) -> Pose:
    """
    :param robot_id:
    :param end_effector:
    :return: Global pose
    """

    robot_inst = await get_robot_instance(robot_id, end_effector)

    if isinstance(robot_inst, Robot):

        try:
            return await hlp.run_in_executor(robot_inst.get_end_effector_pose, end_effector)
        except NotImplementedError:
            raise RobotPoseException("The robot does not support getting pose.")

    elif isinstance(robot_inst, RobotService):

        try:
            return await hlp.run_in_executor(robot_inst.get_end_effector_pose, robot_id, end_effector)
        except NotImplementedError:
            raise RobotPoseException("The robot does not support getting pose.")

    else:
        raise Arcor2Exception("Not a robot instance.")


@scene_needed
@project_needed
async def update_ap_joints_cb(req: rpc.objects.UpdateActionPointJointsRequest) -> \
        Union[rpc.objects.UpdateActionPointJointsResponse, hlp.RPC_RETURN_TYPES]:

    assert SCENE and PROJECT

    try:
        proj_obj, ap = get_object_ap(PROJECT, req.args.id)
    except ActionPointNotFound:
        return False, "Invalid action point."

    try:
        new_joints = await get_robot_joints(req.args.robot_id)
    except Arcor2Exception as e:
        return False, str(e)

    for orientation in ap.orientations:
        if orientation.id == req.args.joints_id:
            return False, "Can't update joints that are paired with orientation."

    for joint in ap.robot_joints:  # update existing joints_id
        if joint.id == req.args.joints_id:
            joint.joints = new_joints
            joint.robot_id = req.args.robot_id
            joint.is_valid = True
            break
    else:
        ap.robot_joints.append(ProjectRobotJoints(req.args.joints_id, req.args.robot_id, new_joints))

    asyncio.ensure_future(notify_project_change_to_others())
    return None


@scene_needed
@project_needed
async def update_action_point_cb(req: rpc.objects.UpdateActionPointPoseRequest) -> \
        Union[rpc.objects.UpdateActionPointPoseResponse, hlp.RPC_RETURN_TYPES]:

    assert SCENE and PROJECT

    try:
        proj_obj, ap = get_object_ap(PROJECT, req.args.id)
    except ActionPointNotFound:
        return False, "Invalid action point."

    try:
        new_pose, new_joints = await asyncio.gather(get_end_effector_pose(req.args.robot.robot_id,
                                                                          req.args.robot.end_effector),
                                                    get_robot_joints(req.args.robot.robot_id))
    except RobotPoseException as e:
        return False, str(e)

    rel_pose = hlp.make_pose_rel(SCENE_OBJECT_INSTANCES[proj_obj.id].pose, new_pose)

    if req.args.update_position:
        ap.position = rel_pose.position

    for ori in ap.orientations:
        if ori.id == req.args.orientation_id:
            ori.orientation = rel_pose.orientation
            break
    else:
        ap.orientations.append(NamedOrientation(req.args.orientation_id, rel_pose.orientation))

    for joint in ap.robot_joints:
        if joint.id == req.args.orientation_id:
            joint.joints = new_joints
            joint.robot_id = req.args.robot.robot_id
            joint.is_valid = True
            break
    else:
        ap.robot_joints.append(ProjectRobotJoints(req.args.orientation_id, req.args.robot.robot_id, new_joints))

    asyncio.ensure_future(notify_project_change_to_others())
    return None


@scene_needed
@no_project
async def update_action_object_cb(req: rpc.objects.UpdateActionObjectPoseRequest) -> \
        Union[rpc.objects.UpdateActionObjectPoseRequest, hlp.RPC_RETURN_TYPES]:

    assert SCENE

    if req.args.id == req.args.robot.robot_id:
        return False, "Robot cannot update its own pose."

    try:
        scene_object = get_scene_object(SCENE, req.args.id)
    except SceneObjectNotFound:
        return False, "Invalid action object."

    try:
        scene_object.pose = await get_end_effector_pose(req.args.robot.robot_id, req.args.robot.end_effector)
    except RobotPoseException as e:
        return False, str(e)

    asyncio.ensure_future(notify_scene_change_to_others())
    return None


def project_problems(scene: Scene, project: Project) -> List[str]:

    scene_objects: Dict[str, str] = {obj.id: obj.type for obj in scene.objects}
    scene_services: Set[str] = {srv.type for srv in scene.services}

    action_ids: Set[str] = set()
    problems: List[str] = []

    unknown_types = ({obj.type for obj in scene.objects} | scene_services) - ACTIONS.keys()

    if unknown_types:
        return [f"Scene invalid, contains unknown types: {unknown_types}."]

    for obj in project.objects:

        # test if all objects exists in scene
        if obj.id not in scene_objects:
            problems.append(f"Object ID {obj.id} does not exist in scene.")
            continue

        for ap in obj.action_points:

            for joints in ap.robot_joints:
                if not joints.is_valid:
                    problems.append(f"Action point {ap.id} has invalid joints: {joints.id} (robot {joints.robot_id}).")

            for action in ap.actions:

                if action.id in action_ids:
                    problems.append(f"Action ID {action.id} of the {obj.id}/{ap.id} is not unique.")

                # check if objects have used actions
                obj_id, action_type = action.parse_type()

                if obj_id not in scene_objects.keys() | scene_services:
                    problems.append(f"Object ID {obj.id} which action is used in {action.id} does not exist in scene.")
                    continue

                try:
                    os_type = scene_objects[obj_id]  # object type
                except KeyError:
                    os_type = obj_id  # service

                for act in ACTIONS[os_type]:
                    if action_type == act.name:
                        break
                else:
                    problems.append(f"Object type {scene_objects[obj_id]} does not have action {action_type} "
                                    f"used in {action.id}.")

                # check object's actions parameters
                action_params: Dict[str, str] = \
                    {param.id: param.type for param in action.parameters}
                ot_params: Dict[str, str] = {param.name: param.type for param in act.parameters
                                             for act in ACTIONS[os_type]}

                if action_params != ot_params:
                    problems.append(f"Action ID {action.id} of type {action.type} has invalid parameters.")

                # TODO validate parameter values / instances (for value) are not available here / how to solve it?
                for param in action.parameters:
                    try:
                        PARAM_PLUGINS[param.type].value(TYPE_DEF_DICT, scene, project, action.id, param.id)
                    except ParameterPluginException:
                        problems.append(f"Parameter {param.id} of action {act.name} "
                                        f"has invalid value: '{param.value}'.")

    return problems


async def list_projects_cb(req: rpc.scene_project.ListProjectsRequest) -> \
        Union[rpc.scene_project.ListProjectsResponse, hlp.RPC_RETURN_TYPES]:

    data: List[rpc.scene_project.ListProjectsResponseData] = []

    projects = await storage.get_projects()

    scenes: Dict[str, Scene] = {}

    for project_iddesc in projects.items:

        try:
            project = await storage.get_project(project_iddesc.id)
        except Arcor2Exception as e:
            await logger.warning(f"Ignoring project {project_iddesc.id} due to error: {e}")
            continue

        pd = rpc.scene_project.ListProjectsResponseData(id=project.id, desc=project.desc)
        data.append(pd)

        if project.scene_id not in scenes:
            try:
                scenes[project.scene_id] = await storage.get_scene(project.scene_id)
            except PersistentStorageException:
                pd.problems.append("Scene does not exist.")
                continue

        pd.problems = project_problems(scenes[project.scene_id], project)
        pd.valid = not pd.problems

        if not pd.valid:
            continue

        try:
            program_src(project, scenes[project.scene_id], otu.built_in_types_names())
            pd.executable = True
        except SourceException as e:
            pd.problems.append(str(e))

    return rpc.scene_project.ListProjectsResponse(data=data)


async def list_scenes_cb(req: rpc.scene_project.ListScenesRequest) -> \
        Union[rpc.scene_project.ListScenesResponse, hlp.RPC_RETURN_TYPES]:

    scenes = await storage.get_scenes()
    return rpc.scene_project.ListScenesResponse(data=scenes.items)


async def list_meshes_cb(req: rpc.storage.ListMeshesRequest) -> Union[rpc.storage.ListMeshesResponse,
                                                                      hlp.RPC_RETURN_TYPES]:
    return rpc.storage.ListMeshesResponse(data=await storage.get_meshes())


async def new_object_type_cb(req: rpc.objects.NewObjectTypeRequest) -> Union[rpc.objects.NewObjectTypeResponse,
                                                                             hlp.RPC_RETURN_TYPES]:

    meta = req.args

    if meta.type in OBJECT_TYPES:
        return False, "Object type already exists."

    if meta.base not in OBJECT_TYPES:
        return False, f"Unknown base object type '{meta.base}', known types are: {', '.join(OBJECT_TYPES.keys())}."

    if not hlp.is_valid_type(meta.type):
        return False, "Object type invalid (should be CamelCase)."

    obj = meta.to_object_type()
    obj.source = new_object_type_source(OBJECT_TYPES[meta.base], meta)

    if meta.object_model and meta.object_model.type != Model3dType.MESH:
        assert meta.type == meta.object_model.model().id
        await storage.put_model(meta.object_model.model())

    # TODO check whether mesh id exists - if so, then use existing mesh, if not, upload a new one
    if meta.object_model and meta.object_model.type == Model3dType.MESH:
        # ...get whole mesh (focus_points) based on mesh id
        assert meta.object_model.mesh
        try:
            meta.object_model.mesh = await storage.get_mesh(meta.object_model.mesh.id)
        except storage.PersistentStorageException as e:
            await logger.error(e)
            return False, f"Mesh ID {meta.object_model.mesh.id} does not exist."

    await storage.update_object_type(obj)

    OBJECT_TYPES[meta.type] = meta
    ACTIONS[meta.type] = otu.object_actions(TYPE_TO_PLUGIN,
                                            hlp.type_def_from_source(obj.source, obj.id, Generic), obj.source)
    otu.add_ancestor_actions(meta.type, ACTIONS, OBJECT_TYPES)

    asyncio.ensure_future(notify(events.ObjectTypesChangedEvent(data=[meta.type])))
    return None


@scene_needed
@no_project
async def focus_object_start_cb(req: rpc.objects.FocusObjectStartRequest) -> Union[rpc.objects.FocusObjectStartResponse,
                                                                                   hlp.RPC_RETURN_TYPES]:

    global FOCUS_OBJECT
    global FOCUS_OBJECT_ROBOT

    obj_id = req.args.object_id

    if obj_id in FOCUS_OBJECT_ROBOT:
        return False, "Focusing already started."

    if obj_id not in SCENE_OBJECT_INSTANCES:
        return False, "Unknown object."

    try:
        inst = await get_robot_instance(req.args.robot.robot_id, req.args.robot.end_effector)
    except Arcor2Exception as e:
        return False, str(e)

    if not ROBOT_META[inst.__class__.__name__].features.focus:
        return False, "Robot/service does not support focusing."

    obj_type = OBJECT_TYPES[get_obj_type_name(obj_id)]

    if not obj_type.object_model or obj_type.object_model.type != Model3dType.MESH:
        return False, "Only available for objects with mesh model."

    assert obj_type.object_model.mesh

    focus_points = obj_type.object_model.mesh.focus_points

    if not focus_points:
        return False, "focusPoints not defined for the mesh."

    FOCUS_OBJECT_ROBOT[req.args.object_id] = req.args.robot
    FOCUS_OBJECT[obj_id] = {}
    await logger.info(f'Start of focusing for {obj_id}.')
    return None


def get_obj_type_name(object_id: str) -> str:

    return SCENE_OBJECT_INSTANCES[object_id].__class__.__name__


@no_project
async def focus_object_cb(req: rpc.objects.FocusObjectRequest) -> Union[rpc.objects.FocusObjectResponse,
                                                                        hlp.RPC_RETURN_TYPES]:

    obj_id = req.args.object_id
    pt_idx = req.args.point_idx

    if obj_id not in SCENE_OBJECT_INSTANCES:
        return False, "Unknown object_id."

    obj_type = OBJECT_TYPES[get_obj_type_name(obj_id)]

    assert obj_type.object_model and obj_type.object_model.mesh

    focus_points = obj_type.object_model.mesh.focus_points

    assert focus_points

    if pt_idx < 0 or pt_idx > len(focus_points)-1:
        return False, "Index out of range."

    if obj_id not in FOCUS_OBJECT:
        await logger.info(f'Start of focusing for {obj_id}.')
        FOCUS_OBJECT[obj_id] = {}

    robot_id, end_effector = FOCUS_OBJECT_ROBOT[obj_id].as_tuple()

    FOCUS_OBJECT[obj_id][pt_idx] = await get_end_effector_pose(robot_id, end_effector)

    r = rpc.objects.FocusObjectResponse()
    r.data.finished_indexes = list(FOCUS_OBJECT[obj_id].keys())
    return r


async def get_robot_instance(robot_id: str, end_effector_id: Optional[str] = None) -> Union[Robot, RobotService]:

    if robot_id in SCENE_OBJECT_INSTANCES:
        robot_inst = SCENE_OBJECT_INSTANCES[robot_id]
        if not isinstance(robot_inst, Robot):
            raise Arcor2Exception("Not a robot.")
        if end_effector_id and end_effector_id not in await hlp.run_in_executor(robot_inst.get_end_effectors_ids):
            raise Arcor2Exception("Unknown end effector ID.")
        return robot_inst
    else:
        robot_srv_inst = find_robot_service()
        if not robot_srv_inst or robot_id not in await hlp.run_in_executor(robot_srv_inst.get_robot_ids):
            raise Arcor2Exception("Unknown robot ID.")
        if end_effector_id and end_effector_id not in await hlp.run_in_executor(robot_srv_inst.get_end_effectors_ids,
                                                                                robot_id):
            raise Arcor2Exception("Unknown end effector ID.")
        return robot_srv_inst


@scene_needed
@no_project
async def focus_object_done_cb(req: rpc.objects.FocusObjectDoneRequest) -> Union[rpc.objects.FocusObjectDoneResponse,
                                                                                 hlp.RPC_RETURN_TYPES]:

    global FOCUS_OBJECT
    global FOCUS_OBJECT_ROBOT

    obj_id = req.args.id

    if obj_id not in FOCUS_OBJECT:
        return False, "focusObjectStart/focusObject has to be called first."

    obj_type = OBJECT_TYPES[get_obj_type_name(obj_id)]

    assert obj_type.object_model and obj_type.object_model.mesh

    focus_points = obj_type.object_model.mesh.focus_points

    assert focus_points

    if len(FOCUS_OBJECT[obj_id]) < len(focus_points):
        return False, "Not all points were done."

    robot_id, end_effector = FOCUS_OBJECT_ROBOT[obj_id].as_tuple()
    robot_inst = await get_robot_instance(robot_id)
    assert hasattr(robot_inst, "focus")  # mypy does not deal with hasattr

    assert SCENE

    obj = get_scene_object(SCENE, obj_id)

    fp: List[Position] = []
    rp: List[Position] = []

    for idx, pose in FOCUS_OBJECT[obj_id].items():

        fp.append(focus_points[idx].position)
        rp.append(pose.position)

    mfa = MeshFocusAction(fp, rp)

    await logger.debug(f'Attempt to focus for object {obj_id}, data: {mfa}')

    try:
        obj.pose = await hlp.run_in_executor(robot_inst.focus, mfa)  # type: ignore
    except Arcor2Exception as e:
        await logger.error(f"Focus failed with: {e}, mfa: {mfa}.")
        return False, "Focusing failed."

    await logger.info(f"Done focusing for {obj_id}.")

    clean_up_after_focus(obj_id)

    asyncio.ensure_future(notify_scene_change_to_others())
    asyncio.ensure_future(scene_object_pose_updated(SCENE.id, obj.id))
    return None


async def scene_object_pose_updated(scene_id: str, obj_id: str) -> None:

    async for project in projects_using_object(scene_id, obj_id):

        for obj in project.objects:
            if obj.id != obj_id:
                continue
            for ap in obj.action_points:
                for joints in ap.robot_joints:
                    joints.is_valid = False

        await storage.update_project(project)


def clean_up_after_focus(obj_id: str) -> None:

    try:
        del FOCUS_OBJECT[obj_id]
    except KeyError:
        pass

    try:
        del FOCUS_OBJECT_ROBOT[obj_id]
    except KeyError:
        pass


async def register(websocket) -> None:

    await logger.info("Registering new ui")
    INTERFACES.add(websocket)

    await notify_scene(websocket)
    await notify_project(websocket)

    # TODO avoid cast
    resp = cast(rpc.execution.ProjectStateResponse,
                await manager_request(rpc.execution.ProjectStateRequest(id=uuid.uuid4().int)))  # type: ignore

    await asyncio.wait([websocket.send(events.ProjectStateEvent(data=resp.data.project).to_json())])
    if resp.data.action:
        await asyncio.wait([websocket.send(events.ActionStateEvent(data=resp.data.action).to_json())])
    if resp.data.action_args:
        await asyncio.wait([websocket.send(events.CurrentActionEvent(data=resp.data.action_args).to_json())])


async def unregister(websocket) -> None:
    await logger.info("Unregistering ui")  # TODO print out some identifier
    INTERFACES.remove(websocket)


async def collision(obj: Generic,
                    rs: Optional[RobotService] = None, *, add: bool = False, remove: bool = False) -> None:
    """

    :param obj: Instance of the object.
    :param add:
    :param remove:
    :param rs:
    :return:
    """

    assert add ^ remove

    if not obj.collision_model:
        return

    if rs is None:
        rs = find_robot_service()
    if rs:
        try:
            # TODO notify user somehow when something went wrong?
            await hlp.run_in_executor(rs.add_collision if add else rs.remove_collision, obj)
        except Arcor2Exception as e:
            await logger.error(e)


async def add_object_to_scene(obj: SceneObject, add_to_scene=True, srv_obj_ok=False) -> Tuple[bool, str]:
    """

    :param obj:
    :param add_to_scene: Set to false to only create object instance and add its collision model (if any).
    :return:
    """

    assert SCENE

    if obj.type not in OBJECT_TYPES:
        # TODO try to get it from storage
        return False, "Unknown object type."

    obj_meta = OBJECT_TYPES[obj.type]

    if obj_meta.disabled:
        return False, "Object type disabled."

    if srv_obj_ok:  # just for internal usage

        if not obj_meta.needs_services <= valid_service_types().keys():
            return False, "Some of required services is not available."

        if not obj_meta.needs_services <= SERVICES_INSTANCES.keys():
            return False, "Some of required services is not in the scene."

    elif obj_meta.needs_services:
        return False, "Service(s)-based object."

    if obj_meta.abstract:
        return False, "Cannot instantiate abstract type."

    if obj.id in SCENE_OBJECT_INSTANCES or obj.id in SERVICES_INSTANCES:
        return False, "Object/service with that id already exists."

    if not hlp.is_valid_identifier(obj.id):
        return False, "Object ID invalid (should be snake_case)."

    await logger.debug(f"Creating instance {obj.id} ({obj.type}).")

    try:

        if obj.type in otu.built_in_types_names():
            cls = otu.get_built_in_type(obj.type)
        else:
            obj_type = await storage.get_object_type(obj.type)
            cls = hlp.type_def_from_source(obj_type.source, obj_type.id, Generic)

        coll_model: Optional[Models] = None
        if obj_meta.object_model:
            coll_model = obj_meta.object_model.model()

        if not obj_meta.needs_services:
            obj_inst = cls(obj.id, obj.pose, coll_model)
        else:

            srv_args: List[Service] = []

            for name, ttype in get_type_hints(cls.__init__).items():

                # service arguments should be listed first
                if not issubclass(ttype, Service):
                    break

                try:
                    srv_args.append(SERVICES_INSTANCES[ttype.__name__])
                except KeyError:
                    return False, f"Object type {obj.type} has invalid typ annotation in the constructor, " \
                                  f"service {ttype.__name__} not available."

            try:
                obj_inst = cls(*srv_args, obj.id, obj.pose, coll_model)  # type: ignore
            except TypeError as e:
                return False, f"System error ({e})."

        SCENE_OBJECT_INSTANCES[obj.id] = obj_inst

        if add_to_scene:
            SCENE.objects.append(obj)

        await collision(obj_inst, add=True)

    except Arcor2Exception as e:
        await logger.error(e)
        return False, "System error"

    return True, "ok"


async def auto_add_object_to_scene(obj_type_name: str) -> Tuple[bool, str]:

    assert SCENE

    if obj_type_name not in OBJECT_TYPES:
        return False, "Unknown object type."

    if obj_type_name in otu.built_in_types_names():
        return False, "Does not work for built in types."

    obj_meta = OBJECT_TYPES[obj_type_name]

    if not obj_meta.needs_services:
        return False, "Ordinary object."

    if obj_meta.abstract:
        return False, "Cannot instantiate abstract type."

    if not obj_meta.needs_services <= valid_service_types().keys():
        return False, "Some of required services is not available."

    if not obj_meta.needs_services <= SERVICES_INSTANCES.keys():
        return False, "Some of required services is not in the scene."

    try:

        obj_type = await storage.get_object_type(obj_type_name)
        cls = hlp.type_def_from_source(obj_type.source, obj_type.id, Generic)

        args: List[Service] = [SERVICES_INSTANCES[srv_name] for srv_name in obj_meta.needs_services]

        assert hasattr(cls, otu.SERVICES_METHOD_NAME)
        for obj_inst in cls.from_services(*args):  # type: ignore

            assert isinstance(obj_inst, Generic)

            if not hlp.is_valid_identifier(obj_inst.id):
                # TODO add message to response
                await logger.warning(f"Object id {obj_inst.id} invalid.")
                continue

            if obj_inst.id in SCENE_OBJECT_INSTANCES:
                await logger.warning(f"Object id {obj_inst.id} already in scene.")
                continue

            SCENE_OBJECT_INSTANCES[obj_inst.id] = obj_inst
            SCENE.objects.append(obj_inst.scene_object())

            if obj_meta.object_model:
                obj_inst.collision_model = obj_meta.object_model.model()
                await collision(obj_inst, add=True)

    except Arcor2Exception as e:
        await logger.error(e)
        return False, "System error"

    return True, "ok"


@scene_needed
@no_project
async def add_object_to_scene_cb(req: rpc.scene_project.AddObjectToSceneRequest) -> \
        Union[rpc.scene_project.AddObjectToSceneResponse, hlp.RPC_RETURN_TYPES]:

    obj = req.args
    res, msg = await add_object_to_scene(obj)

    if not res:
        return res, msg

    asyncio.ensure_future(notify_scene_change_to_others())
    return None


@scene_needed
@no_project
async def auto_add_object_to_scene_cb(req: rpc.scene_project.AutoAddObjectToSceneRequest) -> \
        Union[rpc.scene_project.AutoAddObjectToSceneResponse, hlp.RPC_RETURN_TYPES]:

    obj = req.args
    res, msg = await auto_add_object_to_scene(obj.type)

    if not res:
        return res, msg

    asyncio.ensure_future(notify_scene_change_to_others())
    return None


def find_robot_service() -> Union[None, RobotService]:

    for srv in SERVICES_INSTANCES.values():
        if isinstance(srv, RobotService):
            return srv
    else:
        return None


async def add_service_to_scene(srv: SceneService) -> Tuple[bool, str]:

    if srv.type not in SERVICE_TYPES:
        return False, "Unknown service type."

    if SERVICE_TYPES[srv.type].disabled:
        return False, "Service is disabled."

    if srv.type in SERVICES_INSTANCES:
        return False, "Service already in scene."

    srv_type = await storage.get_service_type(srv.type)
    cls_def = hlp.type_def_from_source(srv_type.source, srv_type.id, Service)

    if issubclass(cls_def, RobotService) and find_robot_service():
        return False, "Scene might contain only one robot service."

    try:
        srv_inst = await hlp.run_in_executor(cls_def, srv.configuration_id)
    except Arcor2Exception as e:
        await logger.error(e)
        return False, "System error"

    SERVICES_INSTANCES[srv.type] = srv_inst

    if isinstance(srv_inst, RobotService):
        await logger.info("RobotService added, adding collision models to all objects.")
        for obj_inst in SCENE_OBJECT_INSTANCES.values():
            await collision(obj_inst, srv_inst, add=True)

    return True, "ok"


@scene_needed
@no_project
async def add_service_to_scene_cb(req: rpc.scene_project.AddServiceToSceneRequest) ->\
        Union[rpc.scene_project.AddServiceToSceneResponse, hlp.RPC_RETURN_TYPES]:

    assert SCENE

    srv = req.args
    res, msg = await add_service_to_scene(srv)

    if not res:
        return res, msg

    SCENE.services.append(srv)
    asyncio.ensure_future(notify_scene_change_to_others())
    return None


async def projects_using_object(scene_id: str, obj_id: str) -> AsyncIterator[Project]:

    id_list = await storage.get_projects()

    for project_meta in id_list.items:

        project = await storage.get_project(project_meta.id)

        if project.scene_id != scene_id:
            continue

        for obj in project.objects:

            if obj.id == obj_id:
                yield project
                break

            for ap in obj.action_points:
                for action in ap.actions:
                    action_obj_id, _ = action.parse_type()

                    if action_obj_id == obj_id:
                        yield project
                        break


@scene_needed
async def scene_object_usage_request_cb(req: rpc.scene_project.SceneObjectUsageRequest) -> \
        Union[rpc.scene_project.SceneObjectUsageResponse, hlp.RPC_RETURN_TYPES]:
    """
    Works for both services and objects.
    :param req:
    :return:
    """

    assert SCENE

    if not (any(obj.id == req.args.id for obj in SCENE.objects) or
            any(srv.type == req.args.id for srv in SCENE.services)):
        return False, "Unknown ID."

    resp = rpc.scene_project.SceneObjectUsageResponse()

    async for project in projects_using_object(SCENE.id, req.args.id):
        resp.data.add(project.id)

    return resp


@scene_needed
async def action_param_values_cb(req: rpc.objects.ActionParamValuesRequest) -> \
        Union[rpc.objects.ActionParamValuesResponse, hlp.RPC_RETURN_TYPES]:

    inst: Union[None, Service, Generic] = None

    # TODO method to get object/service based on ID
    if req.args.id in SCENE_OBJECT_INSTANCES:
        inst = SCENE_OBJECT_INSTANCES[req.args.id]
    elif req.args.id in SERVICES_INSTANCES:
        inst = SERVICES_INSTANCES[req.args.id]
    else:
        return False, "Unknown ID."

    parent_params = {}

    for pp in req.args.parent_params:
        parent_params[pp.id] = pp.value

    try:
        method_name, required_parent_params = inst.DYNAMIC_PARAMS[req.args.param_id]
    except KeyError:
        return False, "Unknown parameter or values not constrained."

    if parent_params.keys() != required_parent_params:
        return False, "Not all required parent params were given."

    # TODO validate method parameters vs parent_params (check types)?

    resp = rpc.objects.ActionParamValuesResponse()

    try:
        method = getattr(inst, method_name)
    except AttributeError:
        await logger.error(f"Unable to get values for parameter {req.args.param_id}, "
                           f"object/service {inst.id} has no method named {method_name}.")
        return False, "System error."

    # TODO update hlp.run_in_executor to support kwargs
    resp.data = await asyncio.get_event_loop().run_in_executor(None, functools.partial(method, **parent_params))
    return resp


@scene_needed
@no_project
async def remove_from_scene_cb(req: rpc.scene_project.RemoveFromSceneRequest) -> \
        Union[rpc.scene_project.RemoveFromSceneResponse, hlp.RPC_RETURN_TYPES]:

    assert SCENE

    if req.args.id in SCENE_OBJECT_INSTANCES:

        SCENE.objects = [obj for obj in SCENE.objects if obj.id != req.args.id]
        obj_inst = SCENE_OBJECT_INSTANCES[req.args.id]
        await collision(obj_inst, remove=True)
        del SCENE_OBJECT_INSTANCES[req.args.id]

    elif req.args.id in SERVICES_INSTANCES:

        # first check if some object is not using it
        for obj in SCENE.objects:
            if req.args.id in OBJECT_TYPES[obj.type].needs_services:
                return False, f"Object {obj.id} ({obj.type}) relies on the service to be removed: {req.args.id}."

        SCENE.services = [srv for srv in SCENE.services if srv.type != req.args.id]
        del SERVICES_INSTANCES[req.args.id]

    else:
        return False, "Unknown id."

    asyncio.ensure_future(remove_object_references_from_projects(req.args.id))
    asyncio.ensure_future(notify_scene_change_to_others())
    return None


async def get_robot_meta_cb(req: rpc.robot.GetRobotMetaRequest) -> Union[rpc.robot.GetRobotMetaResponse,
                                                                         hlp.RPC_RETURN_TYPES]:

    return rpc.robot.GetRobotMetaResponse(data=list(ROBOT_META.values()))


async def system_info_cb(req: rpc.common.SystemInfoRequest) -> Union[rpc.common.SystemInfoResponse,
                                                                     hlp.RPC_RETURN_TYPES]:

    resp = rpc.common.SystemInfoResponse()
    resp.data.version = arcor2.version()
    resp.data.api_version = arcor2.api_version()
    resp.data.supported_parameter_types = set(PARAM_PLUGINS.keys())
    return resp


async def build_project_cb(req: rpc.execution.BuildProjectRequest) -> \
        Union[rpc.execution.BuildProjectResponse, hlp.RPC_RETURN_TYPES]:

    # call build service
    # TODO store data in memory
    with tempfile.TemporaryDirectory() as tmpdirname:

        path = os.path.join(tmpdirname, "publish.zip")

        try:
            await hlp.run_in_executor(rest.download, f"{BUILDER_URL}/project/{req.args.id}/publish", path)
        except rest.RestException as e:
            await logger.error(e)
            return False, "Failed to get project package."

        with open(path, "rb") as zip_file:
            b64_bytes = base64.b64encode(zip_file.read())
            b64_str = b64_bytes.decode()

    # send data to execution service
    exe_req = rpc.execution.UploadPackageRequest(uuid.uuid4().int,
                                                 args=rpc.execution.UploadPackageArgs(req.args.id, b64_str))
    resp = await manager_request(exe_req)
    return resp.result, " ".join(resp.messages) if resp.messages else ""


@scene_needed
@project_needed
async def execute_action_cb(req: rpc.scene_project.ExecuteActionRequest) -> \
        Union[rpc.scene_project.ExecuteActionResponse, hlp.RPC_RETURN_TYPES]:

    assert SCENE and PROJECT

    global RUNNING_ACTION

    if RUNNING_ACTION:
        return False, f"Action {RUNNING_ACTION} is being executed. Only one action can be executed at a time."

    try:
        action = PROJECT.action(req.args.action_id)
    except Arcor2Exception:
        return False, "Unknown action."

    params: Dict[str, Any] = {}

    for param in action.parameters:
        try:
            params[param.id] = PARAM_PLUGINS[param.type].value(TYPE_DEF_DICT, SCENE, PROJECT, action.id, param.id)
        except ParameterPluginException as e:
            await logger.error(e)
            return False, f"Failed to get value for parameter {param.id}."

    obj_id, action_name = action.parse_type()

    obj: Optional[Union[Generic, Service]] = None

    if obj_id in SCENE_OBJECT_INSTANCES:
        obj = SCENE_OBJECT_INSTANCES[obj_id]
    elif obj_id in SERVICES_INSTANCES:
        obj = SERVICES_INSTANCES[obj_id]
    else:
        return False, "Internal error: project not in sync with scene."

    if not hasattr(obj, action_name):
        return False, "Internal error: object does not have the requested method."

    RUNNING_ACTION = action.id

    # schedule execution and return success
    asyncio.ensure_future(execute_action(getattr(obj, action_name), params))
    return None


async def execute_action(action_method: Callable, params: Dict[str, Any]) -> None:

    global RUNNING_ACTION

    assert RUNNING_ACTION

    evt = events.ActionResultEvent()
    evt.data.action_id = RUNNING_ACTION

    try:
        action_result = await hlp.run_in_executor(action_method, *params.values())
    except (Arcor2Exception, TypeError) as e:
        await logger.error(e)
        evt.data.error = str(e)
    else:
        if action_result is not None:
            try:
                evt.data.result = TYPE_TO_PLUGIN[type(action_result)].value_to_json(action_result)
            except KeyError:
                # temporal workaround for unsupported types
                evt.data.result = str(action_result)

    await notify(evt)
    RUNNING_ACTION = None


async def remove_object_references_from_projects(obj_id: str) -> None:

    assert SCENE

    updated_project_ids: Set[str] = set()

    async for project in projects_using_object(SCENE.id, obj_id):

        # delete object and its action points
        project.objects = [obj for obj in project.objects if obj.id != obj_id]

        action_ids: Set[str] = set()

        for obj in project.objects:
            for ap in obj.action_points:

                # delete actions using the object
                ap.actions = [act for act in ap.actions if act.parse_type()[0] != obj_id]

                # delete actions using obj's action points as parameters
                # TODO fix this!
                """
                actions_using_invalid_param: Set[str] = \
                    {act.id for act in ap.actions for param in act.parameters
                     if param.type in (ActionParameterTypeEnum.JOINTS, ActionParameterTypeEnum.POSE) and
                     param.value.startswith(obj_id)}

                ap.actions = [act for act in ap.actions if act.id not in actions_using_invalid_param]

                # get IDs of remaining actions
                action_ids.update({act.id for act in ap.actions})
                """

        valid_ids: Set[str] = action_ids | ActionIOEnum.set()

        # remove invalid inputs/outputs
        for obj in project.objects:
            for ap in obj.action_points:
                for act in ap.actions:
                    act.inputs = [input for input in act.inputs if input.default in valid_ids]
                    act.outputs = [output for output in act.outputs if output.default in valid_ids]

        await storage.update_project(project)
        updated_project_ids.add(project.id)

    await logger.info("Updated projects: {}".format(updated_project_ids))


async def scene_change(ui, event: events.SceneChangedEvent) -> None:

    global SCENE

    if PROJECT:
        await logger.warning("Scene changes not allowed when editing project.")
        return

    if event.data:
        for srv in event.data.services:
            if srv.type not in SERVICES_INSTANCES:
                await notify_scene(ui)
                await logger.warning("Ignoring scene changes: service added.")
                return

        # TODO don't allow change of pose for robots
        for obj in event.data.objects:
            if obj.id not in SCENE_OBJECT_INSTANCES:
                await notify_scene(ui)
                await logger.warning("Ignoring scene changes: object added.")
                return

        SCENE = event.data
    else:
        await clear_scene()

    await notify_scene_change_to_others(ui)


async def project_change(ui, event: events.ProjectChangedEvent) -> None:

    global PROJECT

    # TODO check if all types/actions are valid (not disabled) or make RPC for all changes?
    PROJECT = event.data

    await notify_project_change_to_others(ui)


RPC_DICT: hlp.RPC_DICT_TYPE = {
    rpc.common.SystemInfoRequest: system_info_cb,
    rpc.execution.BuildProjectRequest: build_project_cb,
    rpc.objects.GetObjectTypesRequest: get_object_types_cb,
    rpc.objects.GetActionsRequest: get_object_actions_cb,
    rpc.objects.UpdateActionPointPoseRequest: update_action_point_cb,
    rpc.objects.UpdateActionPointJointsRequest: update_ap_joints_cb,
    rpc.objects.UpdateActionObjectPoseRequest: update_action_object_cb,
    rpc.objects.NewObjectTypeRequest: new_object_type_cb,
    rpc.objects.FocusObjectRequest: focus_object_cb,
    rpc.objects.FocusObjectStartRequest: focus_object_start_cb,
    rpc.objects.FocusObjectDoneRequest: focus_object_done_cb,
    rpc.objects.ActionParamValuesRequest: action_param_values_cb,
    rpc.robot.GetRobotMetaRequest: get_robot_meta_cb,
    rpc.scene_project.SaveProjectRequest: save_project_cb,
    rpc.scene_project.SaveSceneRequest: save_scene_cb,
    rpc.scene_project.OpenProjectRequest: open_project_cb,
    rpc.scene_project.ListProjectsRequest: list_projects_cb,
    rpc.scene_project.ListScenesRequest: list_scenes_cb,
    rpc.scene_project.AddObjectToSceneRequest: add_object_to_scene_cb,
    rpc.scene_project.AutoAddObjectToSceneRequest: auto_add_object_to_scene_cb,
    rpc.scene_project.AddServiceToSceneRequest: add_service_to_scene_cb,
    rpc.scene_project.RemoveFromSceneRequest: remove_from_scene_cb,
    rpc.scene_project.SceneObjectUsageRequest: scene_object_usage_request_cb,
    rpc.scene_project.OpenSceneRequest: open_scene_cb,
    rpc.scene_project.ExecuteActionRequest: execute_action_cb,
    rpc.services.GetServicesRequest: get_services_cb,
    rpc.storage.ListMeshesRequest: list_meshes_cb
}

# add Project Manager RPC API
for k, v in nodes.execution.RPC_DICT.items():

    if v.__name__.startswith("_"):
        continue

    RPC_DICT[k] = manager_request


EVENT_DICT: hlp.EVENT_DICT_TYPE = {
    events.SceneChangedEvent: scene_change,
    events.ProjectChangedEvent: project_change
}


def main():

    assert sys.version_info >= (3, 8)

    parser = argparse.ArgumentParser()

    parser.add_argument("-v", "--verbose", help="Increase output verbosity",
                        action="store_const", const=LogLevel.DEBUG, default=LogLevel.INFO)
    parser.add_argument('--version', action='version', version=arcor2.version(),
                        help="Shows ARCOR2 version and exits.")
    parser.add_argument('--api_version', action='version', version=arcor2.api_version(),
                        help="Shows API version and exits.")
    parser.add_argument("-a", "--asyncio_debug", help="Turn on asyncio debug mode.",
                        action="store_const", const=True, default=False)

    args = parser.parse_args()
    logger.level = args.verbose

    loop = asyncio.get_event_loop()
    loop.set_debug(enabled=args.asyncio_debug)

    loop.run_until_complete(asyncio.wait([asyncio.gather(project_manager_client(), _initialize_server())]))


if __name__ == "__main__":
    main()



class Arcor2Exception(Exception):
    """
    All exceptions are derived from this one.
    """
    pass


class GenericException(Arcor2Exception):
    pass


class RobotException(Arcor2Exception):
    pass


class ResourcesException(Arcor2Exception):
    pass


class ActionPointNotFound(Arcor2Exception):
    pass


class SceneObjectNotFound(Arcor2Exception):
    pass



from arcor2.data.common import Scene, SceneObject
from arcor2.exceptions import SceneObjectNotFound


def get_scene_object(scene: Scene, obj_id: str) -> SceneObject:  # TODO cache it somehow?

    for obj in scene.objects:
        if obj.id == obj_id:
            return obj

    raise SceneObjectNotFound



from dataclasses import dataclass, field
from dataclasses_jsonschema import JsonSchemaMixin


@dataclass
class RobotFeatures(JsonSchemaMixin):

    focus: bool = False


@dataclass
class RobotMeta(JsonSchemaMixin):
    """
    Robot meta that could be extracted without creating an instance.
    """

    type: str
    features: RobotFeatures = field(default_factory=RobotFeatures)



from arcor2.services.service import Service
from arcor2.services.robot_service import RobotService
from arcor2.services.time import TimeService
from arcor2.services.logic import LogicService

__all__ = ["Service", "RobotService", "TimeService", "LogicService"]
